"use strict";

/*
  TRIVIA - Text relates Image VGG Image Annotator (via)

  Copyright (c) 2019, Uwe Sikora, Metadata and Dataconversion Group, Niedersächsische Staats- und Universitätsbibliothek.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
  Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

// https://medium.com/@mattlag/es6-modules-getting-started-gotchas-2ad154f38e2e


/**################################################################################
 *  Classes
 *################################################################################*/

class App {
  constructor( name, version ){
    this.plugins = {};
    this.name = name;
    this.version = version;
  };

  addPlugin( plugin ){
    this.plugins[ plugin.name ] = plugin;
    this.addPluginPanel( plugin )
  };

  addPluginPanel( plugin ){
    let name = plugin.name.toUpperCase();
    if ( plugin.panel && !(name in VIA_DISPLAY_AREA_CONTENT_NAME) ){
      console.log('Plugin "' + plugin.name + '": Add panel-names to via');
      VIA_DISPLAY_AREA_CONTENT_NAME[ name ] = plugin.panel;
    }
  };
}


class Plugin {
  constructor( name, selector ){
    this.name = name;
    this.$element = $( selector );
  };
}


class PanelPlugin extends Plugin {
  constructor( name, selector, panel ) {
    super( name, selector );
    this.panel = panel;
  }

  showPanel(){
    if( !this.panel && !VIA_DISPLAY_AREA_CONTENT_NAME[ name ] ) return;
    let name = this.name.toUpperCase();
    let element = VIA_DISPLAY_AREA_CONTENT_NAME[ name ]
    set_display_area_content( element )
    return $( element ) 
  };
}

/***********************************************************************************
 * Trivia Class
 */
class Trivia extends App {
    constructor( name, version, status = 'dev'){
      super( name, version );
      this.surfaces = new Object()
      this.backend = status != 'dev' ? 'https://dev.ilanot.de.dariah.eu/exist/' : 'https://127.0.1.1:8443/exist/';
      this.world = status == 'dev' ? 'local' : 'living';
    }

    /**
     * Trivia.init()
     * @summary function to init TRIVIA
     * @version 0.1 2019-05-10 stable
     * */
    init(){
      VIA_THEME_SEL_REGION_FILL_BOUNDARY_COLOR = "#0066ff";
      VIA_THEME_BOUNDARY_FILL_COLOR   = "#0066ff";
      VIA_THEME_CONTROL_POINT_COLOR   = '#66ffff';
      
      // 0) Rewrite of image panel (dragscroll)
      VIA_DISPLAY_AREA_CONTENT_NAME['IMAGE'] = "image_panel_frame";
      $('#image_panel_frame').css('height', (window.innerHeight - 100)+'px');
      VIA_REGION_SHAPE['ARROW'] = 'arrow';
      window.addEventListener('resize', console.log(window.innerHeight));

      // 0) Set Zoomlevels
      this.addZoomLevels( [7, 12, 15, 20] );

      let target = this;
        
      // 1.1) VIA Mod.1: installing Wrapperfunctions to modify VIA from Trivia
      window.annotation_editor_show = modAnnotationEditorShow( annotation_editor_show, this )
      window.clone_image_region = modCloneImageRegion( clone_image_region )
      window.select_region_shape = modSelectRegionShape( select_region_shape )

      // 1.2) VIA Mod.2: shadowing attributes editor
      $( "#attributes_editor_panel_title" ).addClass("display_none")

      // 2) init observer on buffer list to update trivias surfaces
      let bufferList = document.getElementById('img_fn_list');
      if( bufferList ) {
        console.log("TRIVIA: Add observer on buffer list");
        
        let bufferListObserver = new MutationObserver(function(mutations) {
            mutations.forEach(function(mutation) {
                target.synch()
            });
        });

        bufferListObserver.observe( bufferList, {
          attributes: false,
          characterData: false,
          childList: true,
          subtree: false,
          attributeOldValue: false,
          characterDataOldValue: false
        });
      }

    }
    
    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * TRIVIA getter
     * **************
     * - activeSurface
     * - viaObjects ( objects from _via_img_metadata )
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/  

    /**
     * Trivia.activeSurface()
     * @summary getter method to get the currently active Surface from TRIVIA
     * @version 0.1 2019-05-10 stable
     * @returns a Trivia.Surface instance
     **/
    get activeSurface() {
      return this.surfaces[ _via_image_id ]
    }
    
    /**
     * Trivia.viaObjects()
     * @summary function to load all data from VIA data variable
     * @version 0.1 2019-05-10 stable
     * @returns an array including all VIA objects
     **/
    get viaObjects() {
      return _via_img_metadata
    };


    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * TRIVIA methods
     * **************
     * - downloadTriviaSurfaces 
     * - getViaObjectById
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/

    /**
     * Trivia.addSurfaceToTrivia()
     * @summary function to load data into VIA
     * @version 0.1 2019-05-10 stable
     * */
    addSurfaceToTrivia( id, object ){
      console.log('Trivia.addSurfaceToTrivia: add surface ' + id + ' into VIA and TriVIA'); 
      let path = object.filename != '' ? object.filename : object.file_attributes.image_url; 

      // update via data structure
      if ( ! _via_img_metadata.hasOwnProperty( id ) ) {
        _via_img_metadata[ id ] = object;
        _via_image_id_list.push( id );
        _via_image_filename_list.push( path );
        _via_img_count += 1;
      }

      //console.log( _via_img_metadata )
      this.surfaces[ id ] = new Surface( id );
    }
    
    /**
     * Trivia.bufferedImages()
     * @summary function to get all buffered images as an object
     * @version 0.1 2019-05-10 stable
     * @returns an object containing all ids and buffered image elements
     * */
    bufferedImages(){
      let bufferObjects = {}
      for ( let i = 0; i < _via_image_id_list.length; i++ ) {
        let bufferedObject = document.getElementById( "bim" + String(i) );
        bufferObjects[ _via_image_id_list[i] ] = bufferedObject;
      };
      return bufferObjects
    };
    
    changeColor( color = "#0066ff" ){
      VIA_THEME_SEL_REGION_FILL_BOUNDARY_COLOR = color;
      VIA_THEME_BOUNDARY_FILL_COLOR   = color;
    }

    /**
     * Trivia.downloadTriviaSurfaces()
     * @summary function to download trivia data
     * @version 0.1 2019-05-10 stable
     * */
    downloadTriviaSurfaces(){
      return JSON.stringify( this.exportAllSurfaces() );
    };
    
    /**
     * Trivia.exportSurface()
     * @summary function to export a surface as an object representation
     * @version 0.1 2019-05-10 stable
     * @returns a object representation of a trivia Surface
     * */
    exportSurface( surfaceId ){
      return this.surfaces[ surfaceId ].export()
    }

    /**
     * Trivia.exportAllSurfaces()
     * @summary function to export all surfaces as an object representation
     * @version 0.1 2019-05-10 stable
     * @returns a object representation of all trivia Surfaces
     * */
    exportAllSurfaces(){
      let objects = {};
      for (let key in this.surfaces){
        objects[key] = this.exportSurface( key )
      }
      return objects
    }

    /**
     * Trivia.getSurfaceSizeByImageId()
     * @summary function to estimate the size of an image
     * @version 0.1 2019-05-10 stable
     * @returns an object containing the natural height and width
     * */
    getSurfaceSizeByImageId( surfaceId ){
      let buffered = this.bufferedImages()[ surfaceId ]
      let naturalHeight = buffered.naturalHeight;
      let naturalWidth = buffered.naturalWidth;
      return { "naturalHeight" : naturalHeight, "naturalWidth" : naturalWidth }
    };

    /**
     * Trivia.getViaObjectByFilename()
     * @summary function to select a VIA object by filename
     * @version 0.1 2019-05-10 stable
     * @returns a via object
     * */
    getViaObjectByFilename( filename ) {
      return this.getViaObjectById( this.filenameToId( filename ) )
    };

    /**
     * Trivia.getViaObjectById()
     * @summary function to select a VIA object by ID
     * @version 0.1 2019-05-10 stable
     * @returns a via object
     * */
    getViaObjectById( id ) {
      return this.viaObjects[ id ]
    };

    /**
     * Trivia.filenameToId()
     * @summary function to map via object filenames to via object ids
     * @version 0.1 2019-05-10 stable
     * @returns a filename : id map
     * */
    filenameToId ( filename ){
      let fileNameIndex = {};
      for (let key in _via_img_metadata) {
          fileNameIndex[ trivia.getViaObjectById(key).filename ] = key
      }
      if ( !filename ) {
        return fileNameIndex
      } else {
        return fileNameIndex[filename]
      } 
    };

    /**
     * Trivia.removeSurfaceFromTrivia()
     * @summary function to remove data from trivia
     * @version 0.1 2019-05-10 stable but dirty
     * function _via_img_buffer_add_image() is a somewhat prototype
     * */
    removeSurfaceFromTrivia( id ){

      // remove image from buffer
      $( this.surfaces[ id ].image ).remove()

      let imageIndex = _via_image_id_list.indexOf( id );
      
      // copy of the VIA project_remove_file() function ################################
      if ( imageIndex < 0 || imageIndex >= _via_img_count ) {
        console.log('project_remove_file(): invalid img_index ' + imageIndex);
        return;
      }
    
      // remove img_index from all array
      // this invalidates all image_index > img_index
      _via_image_id_list.splice( imageIndex, 1 );
      _via_image_filename_list.splice( imageIndex, 1 );
    
      var img_fn_list_index = _via_img_fn_list_img_index_list.indexOf(imageIndex);
      if ( img_fn_list_index !== -1 ) {
        _via_img_fn_list_img_index_list.splice( img_fn_list_index, 1 );
      }
    
      delete _via_img_metadata[ id ];
      delete _via_img_src[ id ];
      delete _via_img_fileref[ id ];
    
      _via_img_count -= 1;
      // END of copy of the VIA project_remove_file() function ########################


      _via_current_image_loaded = false;
      _via_buffer_img_index_list.splice( imageIndex, 1);
      _via_buffer_img_shown_timestamp.splice(imageIndex, 1);
      _via_buffer_img_index_list.splice( imageIndex, 1 )
      
      delete this.surfaces[ id ];
      this.plugins.editor.activeZoneUI = null;
    }

    /**
     * Trivia.synch()
     * @summary function to synchronise the VIA and TRIVIA data
     * @version 0.2 2019-05-21 stable but dirty
     * */
    synch(){
        for( let key in this.viaObjects ){
            if (!(key in this.surfaces)) {
              let surface = new Surface( key )
              surface.addToIndex( this.surfaces )
            }
        }
        for (let key in this.surfaces){
            if ( !( key in this.viaObjects ) ) {
                delete this.surfaces[key]
            }
        }
    }

    addZoomLevels( levels ){
      if ( levels.length ) {
        for ( let i = 0; i < levels.length; i++ ){
          let level = levels[i];
          VIA_CANVAS_ZOOM_LEVELS.push( level )
        }
      }
    };

};


/***********************************************************************************
 * Trivia Surface Class
 * 
 * @description:
 * ...
 * 
 * @todo:
 * - needs a little bit of tidying 
 * - naturalHeight & naturalWidth sollen direkt beim load in die file attributes des via objects geschrieben werden
 * - imgRef & textRef in die file_attributes eines surfaces schreiben
 * - (1) rename to zoneHierarchy, (2) backend: check if i need to den regions and zones at update
 */
class Surface {
    constructor( id ) {
        this.id = id;
        this.viaObject = trivia.getViaObjectById( this.id )
        this.zoneIds = this.viaObject.regions.length
    }

    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * Surface getter
     * **************
     * - canvas (the VIA canvas)
     * - measures
     * - name
     * - image (the VIA Image)
     * - regions
     * - zones
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/  

    /**
     * Surface.canvas
     * @description the canvas of a surface
     * @returns an object
     */
    get canvas(){
      return _via_reg_canvas
    }

    /**
     * Surface.measures
     * @description the absolute height and with of a surfaces image
     * @returns an object
     */
    get measures(){
      return trivia.getSurfaceSizeByImageId( this.id )
    }

    /**
     * Surface.name
     * @description the name of a Trivia Surface
     * @returns an html <img/> dom element
     */
    get name(){
      return this.viaObject.filename
    }

    /**
     * Surface.image
     * @description the buffered image of a surface from the html dom
     * @returns an html <img/> dom element
     */
    get image(){
      return trivia.bufferedImages()[ this.id ]
    }

    /**
     * Surface.regions
     * @description function that returns the list of VIA regions of the surface.
     * @returns a list containing all VIA regions of a surface
     */
    get regions(){
      return this.viaObject.regions
    }

    /**
     * Surface.zones
     * @description function creating corresponding TRIVIA Zone instances from VIA Regions. It also sets a default id if there isn't any, since the zones editor
     *              need ids to operate. Result is an index containing all zone-ids including index from their position in the via objects regions list
     * @returns an object containing zone-ids and corresponding TRIVIA Zone instances 
     */
    get zones() {
      let zonesIndex = {}
      let regions = this.regions
      let index = 0
      for ( let i = 0; i < regions.length; i++ ){
        let region = regions[i]
        let zoneId = region['region_attributes']['zoneId']
        zoneId = zoneId == 'not_defined' || zoneId == 'not defined yet!' || zoneId == undefined || zoneId == '' ? this.newRegionId() : zoneId;
        //console.log(i + ': ' + zoneId)

        let zone = new Zone( this.id, zoneId, region )
        zone.index = index
        zonesIndex[ zoneId ] = zone
        index++
      }
      return zonesIndex
    }

    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * Surface methods
     * **************
     * 
     * - addToVia
     * - addToIndex
     * - export
     * - getRegionByIndex
     * - getRegionById
     * - regionIdNumbers
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/
    
    /**
     * Surface.addToIndex()
     * @description adds a surface to a given surface index object
     * @version 0.1 2019-05-15 stable
     */
    addToIndex( index ){
      if ( !(this.id in index) ){
        index[ this.id ] = this
      } 
    }

    /**
     * Surface.export()
     * @description creates an object representation of a Surface
     * @version 0.1 2019-05-15 stable
     */
    export(){
      let obj = {
        id : this.id,
        filename : '',//this.name,
        size : this.viaObject.size,
        file_attributes : this.viaObject["file_attributes"],
        regions : this.viaObject.regions, 
        zones : this.serialize().zones
      }
      if ( !obj["file_attributes"]["naturalHeight"] || obj["file_attributes"]["naturalHeight"] == '' ) {
        obj["file_attributes"]["naturalHeight"] = this.measures.naturalHeight;
      }
      if ( !obj["file_attributes"]["naturalWidth"] || obj["file_attributes"]["naturalWidth"] == '' ) {
        obj["file_attributes"]["naturalWidth"] = this.measures.naturalWidth;
      }
      obj["file_attributes"]["image_url"] = '';
      return obj
    }

    // Prototyping, experimental
    // tries to minimise the overhead of posted data
    /* export2(){
      let obj = {
        id : this.id,
        filename : '',//this.name,
        size : this.viaObject.size,
        file_attributes : this.viaObject["file_attributes"],
        zones : this.serialize().zones
      }
      if ( !obj["file_attributes"]["naturalHeight"] || obj["file_attributes"]["naturalHeight"] == '' ) {
        obj["file_attributes"]["naturalHeight"] = this.measures.naturalHeight;
      }
      if ( !obj["file_attributes"]["naturalWidth"] || obj["file_attributes"]["naturalWidth"] == '' ) {
        obj["file_attributes"]["naturalWidth"] = this.measures.naturalWidth;
      }
      obj["file_attributes"]["image_url"] = '';
      return obj
    } */

    /**
     * Surface.getRegionByIndex()
     * @description gets a region by its index
     * @version 0.1 2019-05-15 stable
     */
    getRegionByIndex( index ){
      return this.viaObject.regions[ index ]
    }

    /**
     * Surface.getRegionById()
     * @description gets a region by its ID
     * @version 0.1 2019-05-15 stable
     */
    getRegionById( id ){
      //return this.zones[ id ].viaRegion
      return this.zones[ id ]
    }

    /**
     * Surface.regionIdNumbers()
     * @description get all asigned id numbers asigned to the VIA regions
     * @returns the VIA Region object
     */
    regionIdNumbers(){
      let list = []
      for (let key in this.viaObject.regions){
        let id = this.viaObject.regions[key].region_attributes.zoneId;
        let re = /zone(\d+)/
        let extractedId = re.exec(id)
        if( extractedId ){
          list.push( parseInt(extractedId[1]) )
        }
      }
      
      return list.sort( function(a,b){
          return a - b
      })
    }

    /**
     * Surface.removeFromIndex()
     * @description removes a surface from a given surface index object
     * @version 0.1 2019-05-16 stable
     */
    removeFromIndex( index ){
      if ( this.id in index ){
        delete index[ this.id ]
      } 
    }

    /**
     * Surface.updateWithObject()
     * @description function to update the surface data object (aka. the corresponding via _via_image_metadata object) with the given json object
     * @version 0.2 2019-05-10 in_work
     * @summary runtime O(2n)
     * @todo:
     *  - cleaning up: this function iterates to often! Just iterate once and update corresponding VIA Zones
     *  - figuring out how to use the .saved status properly. Mayby during export when the TextGrid update is initialised
     *    from within the editor. so the lastActive list can be checked if saved. if not it should be updated and then pushed to textgrid
     * 
     * @returns the updated regions of a surface
     */
    updateWithObject( zoneList ){
      console.log('Surface.updateWithObject: Updating surface ' + this.id + ' with object from ZoneEditor')
      let zones = this.zones;

      function _updateRegionsFromHierarchy( zoneList ){
        for(let i=0; i < zoneList.length; i++ ){
          let zone = zoneList[i];
          let zoneId = zone.id;
          let children = zone.children;
          
          // update the chilfOf relation of all aggregated zones
          if( children ){
            
            for (let i = 0; i < children.length; i++){
              let source = children[ i ];
              let target = zones[ source.id ];
              target.region_attributes.childOf = zoneId;
            }

            _updateRegionsFromHierarchy( zone.children )
          }
        }
      }

      //console.log( zoneList )
      let regions = this.regions
      
      // clean all region attributes
      for (let key in regions){
        let regionAttr = regions[key]["region_attributes"]
        regionAttr.childOf = "not_defined"
      }

      // update all regions rekursively from the json
      _updateRegionsFromHierarchy( zoneList )
      
      return this.regions
    }

    /**
     * Surface.newRegionId()
     * @description function to create a new region id
     * @returns the new id
     */
    newRegionId(){
      let regionIds = this.regionIdNumbers();
      let newIdNr = regionIds.length == 0 ? 0 : regionIds[ regionIds.length - 1 ] + 1; // get the last id and add 1
      let newId = "zone" + newIdNr;  
      return newId;
    }

    /**
     * Surface.setViaRegionId()
     * @description function to set the zoneId of a VIA Region
     * @returns the VIA Region object
     */
    setViaRegionId( region ){
      region['region_attributes']['zoneId'] = this.newRegionId();
      return region;
    }

    /**
     * Surface.serialize()
     * @description: serializing a surface to a object to build a list from it etc.
     * @version 0.3 2019-05-15 stable 
     * @summary runtime O(n) depending on regions.length
     * @todo: (1) rename to zoneHierarchy
     */
    serialize(){
      let regions = this.regions
      let zones = this.zones
      let rootZones = []
      
      // iterate over Surface.regions in VIA order
      for (let i in regions){
        let region = regions[i];
        let zoneId = region['region_attributes']['zoneId'];
        let childOf = region['region_attributes'].childOf && region['region_attributes'].childOf != "not_defined" ? region['region_attributes'].childOf : null;
          
        // secure if there is no zone with this id: if this zoneid exists if not hang it on to root
        if ( childOf &&  childOf in zones ) { 
          let targetZone = zones[ childOf ];
          
          // if there is no children attribute yet create it and push the current zone into it
          if( !targetZone.children ) { 
            targetZone.children = [] 
          }

          let zone = zones[ zoneId ]
          //zone.data = zone.viaRegion()
          targetZone.children.push( zone.export() )
            
        } else {
          // if it is a root zone that is not aggregated by any othe zone than add it to rootZones, a list containig all root zones
          let zone = zones[ zoneId ]
          //zone.data = zone.viaRegion()
          rootZones.push( zone.export() )
        }
      }

      let obj = { id : this.id, name : this.name, zones : rootZones}
      return obj
    }

};  // END Class "Surface"


/***********************************************************************************
 * @class Zone
 * @description The Zone class encapsulating methods to gather data from via regions and represent them as trivia zones
 * @version 0.1 2019-05-10 stable
 * @param surfaceId the id of the surface a zone is situated on
 * @param zoneId the id of the zone
 */
class Zone {
    constructor( surfaceId, zoneId, viaRegion ) {
      this.surfaceId = surfaceId;
      this.shape_attributes = viaRegion.shape_attributes;
      this.region_attributes = viaRegion.region_attributes;
      this.id = zoneId;
    }

    /**
     * Zone.id Getter
     * @description gets a Zone id
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    get id() {
      return this.region_attributes.zoneId
    }

    /**
     * Zone.id() Setter
     * @description sets a Zone id
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    set id( id ){
      this.region_attributes.zoneId = id;
    }

    get name(){
      return this.region_attributes.name;
    }

    get type(){
      return this.region_attributes.mapItem;
    }

    shape(){
      return this.shape_attributes;
    }

    /**
     * Zone.export()
     * @description exports a Zone
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    export(){
     return this
    }

    

};


/***********************************************************************************
 * Trivia Zone Editor Class returning a editor object
 */
class TriviaZoneEditor extends PanelPlugin {
  constructor( name, panel, element ) {
    super(name, panel, element)
    this.lastActive = null;
    this.activeZoneUI = null;
    this.init()
  }

  /**
   * TriviaZoneEditor.init()
   * @description initiate a Zone Editor instance and installs a save mutation listener on the html element #trivia_zone_pannel
   * @version 0.1 2019-05-10 stable
   * 
   * @param surfaceId the id of a surface
   * @returns a TriviaZoneList instance
   * 
   * @todo:
   *  - the panel id should be given during initialisation of the class
   */
  init(){
    console.log("TRIVIA.ZoneEditor: initiate Zone Editor instance and save mutation listener");
    //let editor = this
    // start listening on trivia_zones_panel
    var targetNode = document.getElementById( this.panel );
    var observer = new MutationObserver(function(){
      if( targetNode.attributes.class.nodeValue == 'display_area_content display_none' && this.activeZoneUI ){
        let status = this.activeZoneUI.saved
        if ( status == false ){
          console.log("TRIVIA.ZoneEditor.saveListener: Save the current zone list on " + this.activeZoneUI.surface.id )
          this.updateSurface( this.activeZoneUI.surface.id )
          //editor.activeZoneUI.saved = true
        } else {
          console.log("TRIVIA.ZoneEditor.saveListener: Zone list of " + this.activeZoneUI.surface.id + " already saved!")
        }
      }
    }.bind(this));
    observer.observe(targetNode, { attributes: true, childList: true });
  }

  /**
   * TriviaZoneEditor.load()
   * @description function to load a zone list from a Surface instance by id
   * @version 0.1 2019-05-10 stable
   * 
   * @param surfaceId the id of a surface
   * @returns a TriviaZoneList instance
   */ 
  load( surfaceId ){
    console.log( "TRIVIA.ZoneEditor.load(): Load zone list on " + surfaceId )
    //let zoneList = this.trivia.surfaces[ surfaceId ].zoneList()
    let zoneUI = this.zoneUI( surfaceId )
    return zoneUI
  }

  /**
   * TriviaZoneEditor.currentList()
   * @description function to load a zone list from the currently active Surface instance
   * @version 0.1 2019-05-10 stable
   * 
   * @returns a TriviaZoneList instance
   */ 
  currentList(){
    let activeSurfaceId = trivia.activeSurface.id
    console.log("TRIVIA.ZoneEditor.currentList(): Start zone editor with list on " + activeSurfaceId + ' as active list')
    this.activeZoneUI = this.load( activeSurfaceId )
    return this.activeZoneUI
  }

  /**
   * TriviaZoneEditor.updateSurface()
   * @description updates a given surface object with the values of the actual Zone Editor
   * @version 0.1 2019-06-18 stable
   * @param surfaceId the ID of the Surface to update
   */
  updateSurface( surfaceId ){
    console.log("TRIVIA.ZoneEditor.updateSurface(): Attempting to update surface " + surfaceId)
    let activeZoneUI = this.activeZoneUI
    let serializedList = activeZoneUI.toObject();
    
    console.log( activeZoneUI )
    console.log( serializedList )
    if ( activeZoneUI.surface && serializedList ){
      console.log("TRIVIA.ZoneEditor.updateSurface(): Updateing surface " + surfaceId)
      activeZoneUI.saved = true;
      return activeZoneUI.surface.updateWithObject( serializedList );
    } else {
      console.log("TRIVIA.ZoneEditor.updateSurface(): ERROR on updating " + surfaceId)
    }
  }

  /**
   * TriviaZoneEditor.toggle()
   * @description toggling function to toggle the html zone editor. It takes the currently active surface and its zoneList, 
   *              replaces the 'old' list with the new one and renders it to HTML. Then it displays the zoneEditor.
   * @verion 0.1 2019-05-10 stable
   */
  toggle() {
    // load editor from currently avtive surface (aka. the image that is shown)
    if ( this.activeZoneUI ){
        this.activeZoneUI.UI.tree('destroy');
    }
    let currentList = this.currentList();
    $('#zones-count').html( '<span class="ids">' + currentList.surface.regions.length + '</span>' );
    $('#image_panel_frame').css('height', (window.innerHeight - 100)+'px');
    this.showPanel();
  };

  /**
   * TriviaZoneEditor.zoneUI()
   * @description creates the zoneList Interface and converts the data into somethin digestible for the jqueryUI
   * @verion 0.1 2019-05-10 stable
   */
  zoneUI( surfaceId ){
    let surface = trivia.surfaces[ surfaceId ];
    let list = [];

    function _render( zone ){
      let obj = { id : zone.id, type : zone.type, index : zone.index }
      obj.name = zone.name == 'not_defined' || zone.name == ''  ? '<'+zone.id+' />' : zone.name;

      if ( !zone.children ){
        return obj
      } else {
        obj.children = [];
        for (let j = 0; j < zone.children.length; j++){
           obj.children.push( _render( zone.children[j] ) )
        }
        return obj
      }
    }

    let hierachy = surface.serialize().zones;
    
    for(let i = 0; i < hierachy.length; i++) {
      list.push( _render(hierachy[i]) )
    } 
    
    let zoneUI = new TriviaZoneUI( surface, this.$element, list );
    return zoneUI.render({
      autoOpen: true,
      dragAndDrop: true,
      buttonLeft: true,
    })
  }

};


class TriviaZoneUI {
  constructor( surface, element, data ) {
    this.surface = surface
    this.data = data;
    this.saved = false;
    this.$element = element; 
    this.UI = null;
  }

  render( config ) {
    config = !config ? { buttonLeft: false, autoOpen: true, dragAndDrop: true } : config;
    config = Object.assign(config, {data : this.data} )
    console.log(config)
    this.UI = this.$element.tree( config );
    return this;
  }

  toJson(){
    return this.UI.tree('toJson')
  }

  toObject(){
    return JSON.parse(this.UI.tree('toJson'))
  }

  convertData( conversion ){
      this.data = conversion( this.data )
      return this.data
  }

  destroy(){
    this.UI.tree('destroy')
    this.UI = null;
    return this
  }
}


/***********************************************************************************
 * TextGrid Class
 * 
 * testuris: textgrid:3r346, textgrid:3r3p5, textgrid:3r527, textgrid:3r6nc
 */
class TriviaTextGridClient extends Plugin {
  constructor( name, element, config ) {
    super(name, element)
    this.config = config;
    this.resources = {};
  }

  /**
   * TriviaTextGridClient.activeResource()
   * @description function to get the currently active resource
   * @returns the currently active resource 
   */
  activeResource(){
    if ( this.resources.length > 0 ){
      return this.resources[ _via_image_id ]
    }
  }

  /**
   * TriviaTextGridClient.init()
   * @description init function
   */
  init( panel ){
    //client.panel = $( panel );
    this.resourceList = this.$element.find('#textgrid_resources_list > ul');

    if( this.config == undefined ){
      this.getConfigFromTrivia()
        .done(function( data ){
          console.log("TriviaTextGridClient.init: Credentials loaded from trivia backend");
          this.config = data;
          this.updatePanel();
        }.bind(this))
        .fail( function( err ){
          console.log("TriviaTextGridClient.init: Error on loading credentials from trivia backend");
        })
    } else {
      console.log("TriviaTextGridClient.init: client not configured");
    }
  }

  /**
   * TriviaTextGridClient.updatePanel()
   * @description function tu update the textgrid project panel 
   */
  updatePanel(){
    this.$element.find( '#trivia-textgrid-project' ).val( this.config.pid );
  }

  /**
   * TriviaTextGridClient.createEntry()
   * @description function to create an entry in textgrid panel list
   * @returns a DOM element 
   */
  createEntry( resource ){
    let resourceList = this.resourceList;
    resourceList.find(".sel").removeClass("sel")
    let listItem = $('<li id="tg'+ (_via_img_count-1) +'" class="trivia-resource-list-item sel" title="Resource Name" data-uri="' + resource.uri + '" data-id="' + resource.uri + '"></li>')
    let title =  $('<span class="trivia-resource-title" onclick="trivia.plugins.tgclient.jumpToImage(this, '+ (_via_img_count-1) +')">' + resource.title + '</span>')
    let icon = $('<span class="trivia-resource-config-icon" onclick="trivia.plugins.tgclient.configTextGridTriviaResource(this)"><svg viewbox="0 0 24 24"><use xlink:href="#icon_settings"></use><title>Configure</title></svg></span>')
    let update = $(' <span class="trivia-resource-config-icon" onclick="trivia.plugins.tgclient.updateTriviaResource(\'' + resource.uri + '\')"><svg viewbox="0 0 24 24"><use xlink:href="#icon_save"></use><title>Update Resource within TextGrid</title></svg></span>')
    listItem.append(title);
    listItem.append(icon);
    listItem.append(update);
    resourceList.append( listItem )
    return listItem
  };

  /**
   * TriviaTextGridClient.configTextGridTriviaResource()
   * @description opens a configuration panel to enter some configuration statements (NOT WORKING)
   * @returns a DOM element 
   * */
  configTextGridTriviaResource( element ){
    let parent = element.parentElement
    let uri = parent.uri
  
    var config = {'title':'Configure the TextGrid TRIVIA Resource Relations' };
    var input = { 
      'trivia-uri': { type:'text', name:'TRIVIA TextGrid Resource (textgrid URI)', placeholder:'e.g. textgrid:1234', disabled:false, size:50 },
      'tei-uri': { type:'text', name:'TEI TextGrid Resource (textgrid URI)', placeholder:'e.g. textgrid:1234', disabled:false, size:50 },
    };
    
    invoke_with_user_inputs(function(){
      console.log(input)
    }, input, config);
  }

  /**
   * TriviaTextGridClient.loading()
   * @description loadingHandler to handle the loading img
   * */
  loading( action ){
    let display = action == 'start' ? 'inline' : 'none';
    $( '.loading-gif' ).css( 'display', display)
  }
  
  /**
   * TriviaTextGridClient.getConfigFromTrivia()
   * @description function to load configured data from the backend
   */
  getConfigFromTrivia(){
    return $.ajax({
      url: trivia.backend + "restxq/trivia/1.1/auth",
      //url: 'http://localhost:8080' + "restxq/trivia/1.1/auth",
      type: 'GET',
      crossDomain: true,
      //dataType : "json"
    });
  }

  /* testFetch(){
    fetch( trivia.backend +  'restxq/trivia/1.1/auth')
            .then( data => {return data.json()})
            .then( res => {console.log(res)})
  } */

  /**
   * TriviaTextGridClient.loadTriviaResource()
   * @description function to load a trivia resource from textgrid and backend into trivia frontend using a via user input form.
   *              This is the main interface to load trivia resources into the frontend
   * @version 0.1 2019-05-10 working 
   */
  loadTriviaResource(){
    console.log('TriviaTextGridClient.loadTriviaResource: Loading trivia resource into trivia frontend')
    
    let client = this;
    let config = { 'title':'Load Trivia Resource from TextGrid' };
    let input = { 
      'textGridUri': { type:'text', name:'TextGrid URI of a Trivia Resource', placeholder:'textgrid:3r3p5', disabled:false, size:50 },
      };
  
    let loadHandle = function(){
      if ( input.textGridUri.value !== '' ) {
        let uri = input.textGridUri.value.trim();
        console.log('TriviaTextGridClient.loadTriviaResource.loadHandle: Retriving ' + uri + ' from textgrid ...')
        
        let triviaResource = new TriviaResource( uri )
        triviaResource.set( this.config )
        //triviaResource.register( client )
        this.loading( 'start' )
        
        triviaResource.ingest()
          .done( function() { 
            this.loading( 'stop' )
            triviaResource.listElement = this.createEntry( triviaResource )
            triviaResource.available = true;
            this.resources[ triviaResource.uri ] = triviaResource
            show_message('Added TextGrid resource ' + triviaResource.uri + ' with image resource [' + triviaResource.imageRef + ']');
          }.bind(this))
          .fail( function(){
            console.log( "TriviaTextGridClient.loadTriviaResource.loadHandle: Failed to load Trivia Resource" )
            show_message( 'Error: Failed to load TextGrid resource ' + triviaResource.uri );
            triviaResource.available = false;
            this.loading( 'stop' )
          }.bind(this))
      }
    }.bind(this);

    invoke_with_user_inputs( loadHandle, input, config );
  }

  /**
   * TriviaTextGridClient.createTriviaResource()
   * @description function to create a trivia resource within textgrid using a via user input form.
   *              This is the main interface to create trivia resources in textgrid
   * @version 0.1 2019-05-10 in development 
   * @todo:
   *  - if image uri is not in textgrid dont create > call tgcrud for image metadata in frontend
   */
  createTriviaResource(){
    console.log('TriviaTextGridClient.createTriviaResource: Creating trivia resource within textgrid')
    
    let client = this;
    let config = { 'title':'Create Trivia Resource from TextGrid' };
    let input = { 
      'title'   : { type:'text', name:'Title of Resource', placeholder:'The Title of the created Trivia Resource eg. trivia resource' +  Date.now(), disabled:false, size:50 },
      'imageUri': { type:'text', name:'Image Resource', placeholder:' The textgrid uri of the image resource eg. textgrid:3r3hn', disabled:false, size:50 },
      'textUri': { type:'text', name:'TEI Resource', placeholder:' The textgrid uri of an associated TEI resource eg. textgrid:3r3hn', disabled:false, size:50 },
      };
  
    let createHandle = function(){
      let title = input.title.value.trim();
      let imageUri = input.imageUri.value.trim();
      let textUri = input.textUri.value.trim();

      if ( title !== '' && imageUri !== '' ) {
        console.log('TriviaTextGridClient.createTriviaResource: Creating resource [' + title + '] in textgrid');
        let triviaResource = new TriviaResource()
        //triviaResource.register( client )
        triviaResource.set( client.config )
        client.loading( 'start' )
        
        triviaResource.create( title, imageUri, textUri )
          .done( function(){ 
            triviaResource.ingest()
              .done( function () {
                triviaResource.listElement = client.createEntry( triviaResource )
                triviaResource.available = true;
                client.resources[ triviaResource.uri ] = triviaResource
                client.loading( 'stop' )
                show_message('Created and Added TextGrid resource ' + triviaResource.uri + ' with image resource [' + triviaResource.imageRef + ']');
              })
          })
          .fail( function(){
            console.log( "TriviaTextGridClient.createTriviaResource.createHandle: Failed to create Trivia Resource" )
            show_message( 'Error: Failed to create new TextGrid resource!' );
            triviaResource.available = false;
            client.loading( 'stop' )
          })
      } else {
        console.log('TriviaTextGridClient.createTriviaResource: Could not create Trivia Resource without Title or Image URI!');
        show_message('Error: Could not create Trivia Resource without Title or Image URI!!');
      }
    };

    invoke_with_user_inputs( createHandle, input, config );
  }

  /**
   * TriviaTextGridClient.removeTriviaResource()
   * @description function to remove a trivia resource from the frontend
   *              This is the main interface to remove trivia resources
   * @version 0.1 2019-05-10 development 
   * @todo:
   *  - implement functionality
   */
  removeTriviaResource(){
    console.log('TriviaTextGridClient.removeTriviaResource: Removing trivia resource from trivia frontend')
    let id = trivia.activeSurface.id;
    trivia.removeSurfaceFromTrivia( id );
    // not nice but handy to load the first image off all
    if ( Object.keys(_via_img_metadata).length > 0 ){
      _via_show_img(0)
    } else {
      set_display_area_content(VIA_DISPLAY_AREA_CONTENT_NAME.PAGE_GETTING_STARTED);
    }
    
    $( this.resources[ id ].listElement ).remove()
    show_message('Deleted TextGrid resource ' + id + ' from TriVIA');
  }

  /**
   * TriviaTextGridClient.updateTriviaResource()
   * @description function to update a trivia resource in TextGrid
   *              This is the main interface to update trivia resources
   * @version 0.1 2019-05-10 development 
   * @todo:
   *  - implement functionality
   */
  updateTriviaResource( uri ){
    console.log('TriviaTextGridClient.updateTriviaResource: Updateing Trivia Resource in TextGrid')
    this.loading( 'start' )
    
    console.log(trivia.plugins.editor)
    //console.log( editor.activeZoneUI.saved );
    if( trivia.plugins.editor.activeZoneUI && trivia.plugins.editor.activeZoneUI.saved != true ){
      console.log("Not saved yet! needed implementation");
      trivia.plugins.editor.updateSurface( uri );
      // update current editor
    }

    console.log('AJAX request update');
    this.resources[ uri ].update()
    .done( function( data ) { 
      console.log("Ajax Request done")
      show_message('Updated TextGrid Resource ' + uri);
      this.loading( 'stop' )
    }.bind( this ))
    .fail( function( err ){
      console.log( "Error: Ajax update failed" );
      console.log( err )
      show_message('Error on updating TextGrid Resource ' + uri);
      this.loading( 'stop' )
    }.bind( this ));
  }

  jumpToImage( element, index ){
    $("#textgrid_resources_list > ul").children().removeClass("sel")
    let parent = $( element.parentElement ).addClass('sel');
    _via_show_img(index);
    //jump_to_image(index)
  }
    
};


/***********************************************************************************
 * TextGridResource
 */
class TextGridResource{
  constructor( uri, sid, pid, crud = 'https://textgridlab.org/1.0/tgcrud/rest' ){
    this.uri = uri;
    this.sid = sid;
    this.pid = pid;
    this.crud = crud;
  }

  /**
   * TextGridResource.register()
   * @description registering the TextGridResource instance in the given TriviaTextGridClient Instance and loading config
   * @param tgclient the TriviaTextGridClient Instance this resource should be registerd in
   * @returns itself (instance of TextGrid Resource)
   */
  register( tgclient ){
    this.set( tgclient.config )
    tgclient.resources[ this.uri ] = this
    return this
  }

  get( uri ){
    console.log( "get metadata and data from textgrid" )
    // nearly the same functionailty as in trivia resource
  };

  /**
   * TextGridResource.set()
   * @description setting the parameters
   * @param config 
   * @todo:
   *  - delete the hard coded credentials and load from somewhere else maybe via tgclient
   */
  set( config ){
    this.pid = config['pid'];
    console.log("localStorage: " + config['sid']);
    this.sid   =   window.localStorage.getItem('sid') ? window.localStorage.getItem('sid') : config['sid'],
    tglog( 'seta: ' + this.sid );
    this.crud  =   config['tgCrud'],
    this.aggregations   =   config['aggregations']
    return this
  }

  /**
   * TextGridResource.meta()
   * @description getting metadata of a textgrid resource from tgcrud
   * @returns an ajax object
   * @todo:
   */
  meta(){
    return $.ajax({
      url: this.crud + '/' + this.uri + '/metadata?sessionId=' + this.sid,
      type: 'GET',
      crossDomain: true,
      contentType: "json",
    })
  }

  /**
   * TextGridResource.data()
   * @description getting a textgrid resource from tgcrud
   * @returns an ajax object
   * @todo:
   */
  data(){
    return $.ajax({
      url: this.crud + '/' + this.uri + '/data?sessionId=' + this.sid,
      type: 'GET',
      crossDomain: true,
      contentType: "json",
    })
  }


};


/***********************************************************************************
 * TriviaResource < TextGridResource
 */
class TriviaResource extends TextGridResource{
  constructor( uri, sid, pid, crud ){
    super(uri, sid, pid, crud)
    this.triviaCrud = trivia.backend + "restxq/trivia/1.1/textgrid/crud";
    this.textRef = [];
    this.imageRef = null;
  };

  /**
   * TriviaResource.create()
   * @summary function to create a trivia resource in textgrid
   * @version 0.1 2019-05-10 in progress
   * @returns a ajax object
   * @todo:
   *  - this function is quick and dirty ... so tidy up!
   * */
  create( title, imageRef, textUri ){
    console.log("TriviaResource.create: create trivia resource")
    let url = this.triviaCrud + '/create?sid=' + this.sid + '&pid=' + this.pid 
    url += '&imageUri=' + imageRef;
    url += textUri ? '&textUri=' + textUri : '';
    url += '&title=' + title;
    url += this.aggregations.trivia ? '&aggregation=' + this.aggregations.trivia : '';

    console.log( encodeURI(url) )
    
    let create = $.ajax({
      url: encodeURI(url),
      type: 'GET',
      crossDomain: true,
      //contentType: "json",
    })

    let instance = this 
    return create.then( function( data ){
      instance.uri = data.created.triviaResource.uri;
      instance.textRef = data.created.teiResource ? data.created.teiResource.uri : ''
    })
    
  }

  /**
   * TriviaResource.data()
   * @summary function to get a converted trivia resource from textgrid via trivia. This function is a trivia wrapper around the textgrid http GET data method.
   * 
   * @returns an jquery ajax object that may be chained
   * */
  data(){
    return $.ajax({
      url: this.triviaCrud + '/' + this.uri + '/data?sid=' + this.sid,
      type: 'GET',
      crossDomain: true,
      //contentType: "json",
    })
  }

  /**
   * TriviaResource.get()
   * @summary function to load data from textgrid into the Trivia Resource instance
   * 
   * @returns a jquery ajax object containing the finished get request
   *  */  
  get(){
    console.log( "TriviaResource.get: get converted recource from textgrid via trivia api" )
    let instance = this

    let metadataObject = this.meta();
    let dataObject  = this.data();
    
    // fire metadata request and then ...
    return metadataObject.then(
      function( metadata ){
        console.log("TriviaResource.get: get title from TextGrid");
        instance.title = metadata.getElementsByTagName("title")[0].innerHTML;
        // console.log( metadata );

        // ... fire data request to exist-db backend to retrieve converted trivia resource
        return dataObject.done( function( data ){
          //dataCallback( data, instance );
          console.log("TriviaResource.get: get data from TextGrid"); 
          // console.log(data);
          instance.object = data;
          instance.imageRef = data.file_attributes.imageRef 
          instance.textRef = data.file_attributes.textRef 
      
      // set image_url in filename and file_attributes.image_url
      let imageUrl = instance.imageUrl();
      instance.object.filename = imageUrl;
      instance.object.file_attributes.image_url = imageUrl;
        })
        .fail( function() {
          console.log( "TriviaResource.get.data: error loading TRIVIA Resource")
        })

      }
    ).fail( function() {
      console.log( "TriviaResource.get.metadata: error loading TRIVIA Resource")
    })
  };

  /**
   * TriviaResource.imageUrl()
   * @summary function to construct a textgrid url for a image resource
   * 
   * @returns an url as string
   * */
  imageUrl( imageUrl ){
    let uri = imageUrl ? imageUrl : this.imageRef
    let url = this.crud + "/" + uri + '/data'
    url += this.sid ? '?sessionId=' + this.sid : ""
    return url 
  }

  /**
   * TriviaResource.ingest()
   * @summary function to "ingest" aka. load data from textgrid and trivia backend to trivia frontend.
   *          It loads the data into the frontend data- and html structure.
   * @todo bind it to surface as method
   * @version 0.1 2019-05-10 stable
   * */
  ingest(){
    console.log("TriviaResource.ingest: Loading trivia resource from textgrid into trivia application");
    
    let instance = this;
    return this.get().done( function(){
      console.log('TriviaResource.ingest: retrived data for ' + instance.uri + ' from TextGrid. Ingesting into Trivia'); 
      trivia.addSurfaceToTrivia( instance.uri, instance.object )
        let img_index = _via_image_id_list.indexOf( instance.uri );
        _via_show_img(img_index);
        trivia.synch();
      })
  };

  /**
   * TriviaResource.update()
   * @summary updates a resource within textgrid using the crud
   * @todo bind it to surface as method
   * @version 0.1 2019-05-10 stable
   * */
  update( ){
    console.log( 'TRIVIA.TriviaResource.update: Updating trivia resource ' + this.uri + ' within textgrid ...');
    let uri = this.uri;
    
    // check if resource is actually represented in surfaces
    if ( trivia.surfaces[ uri ] ) {
      this.object = trivia.surfaces[ uri ].export();
      
      //console.log( this )
      let postData = JSON.stringify( this.object );
      let url = this.triviaCrud + '/' + this.uri + '/update?sid=' + this.sid;
      //let url = 'http://localhost:8080/exist/restxq/trivia/1.3/textgrid/crud/' + this.uri + '/update?sid=' + this.sid;
      
      //console.log( postData );
      console.log( url );  

      
        // var req = new XMLHttpRequest();
        // // Cross Origin for testing purposes
        // req.open('POST', url, true);
        // //req.setRequestHeader('Content-Type', 'application/json');
        // req.onreadystatechange = function() {
        //   if (req.readyState === 4) {
        //       if (req.status >= 200 && req.status < 400) {
        //         console.log("XMLRequest done")
        //         //console.log(req.responseText);
        //         tgclient.loading( 'stop' )
        //         show_message('Updated TextGrid Resource ' + uri);
        //         //return req.responseText;
        //       } else {
        //         console.log( "Error: XMLRequest update failed" );
        //         show_message('Error on updating TextGrid Resource ' + uri);
        //         tgclient.loading( 'stop' )
        //         console.log(req)
        //       }
        //     }
        //   };
        // req.send( postData );

        return $.ajax({
          url: url,
          type: 'POST',
          crossDomain: true,
          data: postData,
          dataType: "json",
          contentType: "application/json",
        })

    } else {
      console.log('TRIVIA.TriviaResource.update: Error since ' + uri + ' is not a surface!' )
    }
  }

};


/**################################################################################
 * VIA Modifications
 * ~~~~~~~~~~~~~~~~~
 * - Wrapperfunction 'annotation_editor_show'
 * - Wrapperfunction 'clone_image_region'
 *         
 *###############################################################################
 */

/**
 * modZoneEditor()
 * @summary modifies the VIA Region Editor
 * @version 0.1 2019-06-18 stable
 * */
var modAnnotationEditorShow = function( fn , trivia ){
  return function(){
    try{
      
      // build dropdown on attribute childId
      let options = {}
      let regions = trivia.activeSurface.regions;
      let selectedRegionId = _via_user_sel_region_id;
      for (let id = 0; id < regions.length; id++){
        
        if( id != selectedRegionId ){
          let region = regions[ id ];
          let count = id + 1;
          //console.log(region)
          let zoneId = region.region_attributes.zoneId;
          let name = region.region_attributes.name;
          let string = name ? name + ' (' + zoneId + ')' : zoneId;
          options[ zoneId ] = count + ') ' + string;
        }
      }
      options[ 'not_defined' ] = '0) image';
      _via_attributes['region']['childOf'].type = 'dropdown';
      _via_attributes['region']['childOf'].options = options;


      let funcReturn = fn.apply(fn, arguments);

      /*
      $("#annotation_editor").find( $("textarea")[1] ).change( function(){ 
        console.log( trivia.activeSurface().regionIds() )
      })*/
      
    }catch(ex){
      //ErrorHandler.Exception(ex);
      console.log("error on modified function modAnnotationEditorShow ('annotation_editor_show')")
      console.log( ex )
    }
  };
}

/**
 * modZoneEditor()
 * @summary modifies copy functionality that way that IDs are not copied
 * @version 0.1 2019-06-18 stable
 * */
var modCloneImageRegion = function( fn ){
  return function(){
    try{
      let funcReturn = fn.apply(fn, arguments);
      funcReturn.region_attributes.zoneId = undefined;
      return funcReturn;
      
    }catch(ex){
      console.log("error on modified function modCloneImageRegion ('clone_image_region')")
      console.log( ex )
    }
  };
}

/**
 * modZoneEditor()
 * @summary modifies copy functionality that way that IDs are not copied
 * @version 0.1 2019-06-18 stable
 * */
var annotation_editor_get_placement = function(){
  var x = event.clientX + 15;     // Get the horizontal coordinate
  var y = event.clientY + 5;     // Get the vertical coordinate
  var coor = "X coords: " + x + ", Y coords: " + y;
  console.log(coor)
  return {top : y+'px', left : x+'px'}
}

/**
 * modZoneEditor()
 * @summary modifies copy functionality that way that IDs are not copied
 * @version 0.1 2019-06-18 stable
 * */
var modSelectRegionShape = function( fn ){
  return function(){
    try{
      if(arguments[0] === 'arrow') {
        
        for ( var shape_name in VIA_REGION_SHAPE ) {
          var ui_element = document.getElementById('region_shape_' + VIA_REGION_SHAPE[shape_name]);
          ui_element.classList.remove('selected');
        }

        let img = document.getElementById('image_panel_frame');
        let blender = document.getElementById('blender');
        var arrow = document.getElementById('region_shape_arrow');
        let panel = document.getElementById('image_panel');
        
        _via_current_shape = 'pointer';

        console.log(img)
        arrow.classList.add('selected');
        // img.classList.add("dragscroll");
        // blender.style.height = panel.clientHeight + "px";
        // blender.style.width = panel.clientHeight + "px";
        // blender.style.display = 'block';
        console.log('on')
        dragscroll.reset()
        // butClick()
      } else {
        var arrow = document.getElementById('region_shape_arrow');
        let img = document.getElementById('image_panel_frame');
        let blender = document.getElementById('blender');

        if ( img.classList.contains("dragscroll") ) { 
          console.log('off')
          arrow.classList.remove('selected');
          // img.classList.remove("dragscroll");
          // blender.style.display = 'none';
          dragscroll.reset()
        }
        
        let funcReturn = fn.apply(fn, arguments);
        return funcReturn;
      }
      
      
    }catch(ex){
      console.log("error on modified function modSelectRegionShape ('select_region_shape')")
      console.log( ex )
    }
  };
}

function butClick(){
  if ( img.classList.contains("dragscroll") ) {
    console.log('off')
    img.classList.remove("dragscroll");
    blender.style.display = 'none';
  } else {
    console.log('on')
    img.classList.add("dragscroll");
    blender.style.height = panel.clientHeight + "px";
    blender.style.width = panel.clientHeight + "px";
    blender.style.display = 'block';
    console.log(blender)
  }
  dragscroll.reset()
}