/*
  TRIVIA - Text relates Image VGG Image Annotator (via)

  Copyright (c) 2019, Uwe Sikora, Metadata and Dataconversion Group, Niedersächsische Staats- und Universitätsbibliothek.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
  Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

"use strict";


/**################################################################################
 *  Classes
 *################################################################################*/


/***********************************************************************************
 * Trivia Class
 */
class Trivia {
    constructor( status = 'dev'){
        this.surfaces = new Object()
        this.version = '1.0.1';
        this.backend = status != 'dev' ? 'https://dev.ilanot.de.dariah.eu/exist/' : 'https://127.0.1.1:8443/exist/';
        this.name = 'TRIVIA - VGG Image Annotator';
        this.shortName = 'TRIVIA';
        this.panels = { ZONES : 'trivia_zones_panel' }
    }

    /**
     * Trivia.init()
     * @summary function to init TRIVIA
     * @version 0.1 2019-05-10 stable
     * */
    init(){
      VIA_THEME_SEL_REGION_FILL_BOUNDARY_COLOR = "#0066ff";
      VIA_THEME_BOUNDARY_FILL_COLOR   = "#0066ff";
      VIA_THEME_CONTROL_POINT_COLOR   = '#66ffff';

      let target = this;
        
      // 1) VIA Mod.1: Adding new pannel to VIA
      for (let key in this.panels){
          console.log("TRIVIA: Add panel-names to via");
          if (!(key in VIA_DISPLAY_AREA_CONTENT_NAME)){ VIA_DISPLAY_AREA_CONTENT_NAME[key] = this.panels[key] }
      }

      // 1.2) VIA Mod.2: shadowing attributes editor
      $( "#attributes_editor_panel_title" ).addClass("display_none")
      //$( "#attributes_editor_panel" ).empty()
      

      // 2) init observer on buffer list to update trivias surfaces
      let bufferList = document.getElementById('img_fn_list');
      if( bufferList ) {
        console.log("TRIVIA: Add observer on buffer list");
        
        let bufferListObserver = new MutationObserver(function(mutations) {
            mutations.forEach(function(mutation) {
                target.synch()
            });
        });

        bufferListObserver.observe( bufferList, {
          attributes: false,
          characterData: false,
          childList: true,
          subtree: false,
          attributeOldValue: false,
          characterDataOldValue: false
        });
      }
      
    }
    
    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * TRIVIA getter
     * **************
     * - activeSurface
     * - viaObjects ( objects from _via_img_metadata )
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/  

    /**
     * Trivia.activeSurface()
     * @summary getter method to get the currently active Surface from TRIVIA
     * @version 0.1 2019-05-10 stable
     * @returns a Trivia.Surface instance
     **/
    get activeSurface() {
      return this.surfaces[ _via_image_id ]
    }
    
    /**
     * Trivia.viaObjects()
     * @summary function to load all data from VIA data variable
     * @version 0.1 2019-05-10 stable
     * @returns an array including all VIA objects
     **/
    get viaObjects() {
      return _via_img_metadata
    };


    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * TRIVIA methods
     * **************
     * - downloadTriviaSurfaces 
     * - getViaObjectById
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/

    /**
     * Trivia.addSurfaceToTrivia()
     * @summary function to load data into VIA
     * @version 0.1 2019-05-10 stable
     * */
    addSurfaceToTrivia( id, object ){
      //console.log('TriviaResource.addSurfaceToTrivia: add surface ' + id + ' into VIA and TriVIA'); 
      let path = object.filename != '' ? object.filename : object.file_attributes.image_url; 

      // update via data structure
      if ( ! _via_img_metadata.hasOwnProperty( id ) ) {
        _via_img_metadata[ id ] = object;
        _via_image_id_list.push( id );
        _via_image_filename_list.push( path );
        _via_img_count += 1;
      }

      console.log( _via_img_metadata )
      
      this.surfaces[ id ] = new Surface( id );
    }
    

    /**
     * Trivia.bufferedImages()
     * @summary function to get all buffered images as an object
     * @version 0.1 2019-05-10 stable
     * @returns an object containing all ids and buffered image elements
     * */
    bufferedImages(){
      let bufferObjects = {}
      for ( let i = 0; i < _via_image_id_list.length; i++ ) {
        let bufferedObject = document.getElementById( "bim" + String(i) );
        bufferObjects[ _via_image_id_list[i] ] = bufferedObject;
      };
      return bufferObjects
    };
    

    changeColor( color = "#0066ff" ){
      VIA_THEME_SEL_REGION_FILL_BOUNDARY_COLOR = color;
      VIA_THEME_BOUNDARY_FILL_COLOR   = color;
    }

    /**
     * Trivia.downloadTriviaSurfaces()
     * @summary function to download trivia data
     * @version 0.1 2019-05-10 stable
     * */
    downloadTriviaSurfaces(){
      return JSON.stringify( this.exportAllSurfaces() );
    };
    
    
    /**
     * Trivia.exportSurface()
     * @summary function to export a surface as an object representation
     * @version 0.1 2019-05-10 stable
     * @returns a object representation of a trivia Surface
     * */
    exportSurface( surfaceId ){
      return this.surfaces[ surfaceId ].export()
    }


    /**
     * Trivia.exportAllSurfaces()
     * @summary function to export all surfaces as an object representation
     * @version 0.1 2019-05-10 stable
     * @returns a object representation of all trivia Surfaces
     * */
    exportAllSurfaces(){
      let objects = new Object();
      for (let key in this.surfaces){
        objects[key] = this.exportSurface( key )
      }
      return objects
    }

    
    /**
     * Trivia.getSurfaceSizeByImageId()
     * @summary function to estimate the size of an image
     * @version 0.1 2019-05-10 stable
     * @returns an object containing the natural height and width
     * */
    getSurfaceSizeByImageId( surfaceId ){
      let buffered = this.bufferedImages()[ surfaceId ]
      let naturalHeight = buffered.naturalHeight;
      let naturalWidth = buffered.naturalWidth;
      return { "naturalHeight" : naturalHeight, "naturalWidth" : naturalWidth }
    };


    /**
     * Trivia.getViaObjectByFilename()
     * @summary function to select a VIA object by filename
     * @version 0.1 2019-05-10 stable
     * @returns a via object
     * */
    getViaObjectByFilename( filename ) {
      return this.getViaObjectById( this.filenameToId( filename ) )
    };


    /**
     * Trivia.getViaObjectById()
     * @summary function to select a VIA object by ID
     * @version 0.1 2019-05-10 stable
     * @returns a via object
     * */
    getViaObjectById( id ) {
      return this.viaObjects[ id ]
    };


    /**
     * Trivia.filenameToId()
     * @summary function to map via object filenames to via object ids
     * @version 0.1 2019-05-10 stable
     * @returns a filename : id map
     * */
    filenameToId ( filename ){
      let fileNameIndex = new Object();
      for (let key in _via_img_metadata) {
          fileNameIndex[ trivia.getViaObjectById(key).filename ] = key
      }
      if ( !filename ) {
        return fileNameIndex
      } else {
        return fileNameIndex[filename]
      } 
    };


    /**
     * Trivia.removeSurfaceFromTrivia()
     * @summary function to remove data from trivia
     * @version 0.1 2019-05-10 stable but dirty
     * function _via_img_buffer_add_image() is a somewhat prototype
     * */
    removeSurfaceFromTrivia( id ){

      // remove image from buffer
      $( this.surfaces[ id ].image ).remove()

      let imageIndex = _via_image_id_list.indexOf( id );
      
      // copy of the VIA project_remove_file() function ################################
      if ( imageIndex < 0 || imageIndex >= _via_img_count ) {
        console.log('project_remove_file(): invalid img_index ' + imageIndex);
        return;
      }
    
      // remove img_index from all array
      // this invalidates all image_index > img_index
      _via_image_id_list.splice( imageIndex, 1 );
      _via_image_filename_list.splice( imageIndex, 1 );
    
      var img_fn_list_index = _via_img_fn_list_img_index_list.indexOf(imageIndex);
      if ( img_fn_list_index !== -1 ) {
        _via_img_fn_list_img_index_list.splice( img_fn_list_index, 1 );
      }
    
      delete _via_img_metadata[ id ];
      delete _via_img_src[ id ];
      delete _via_img_fileref[ id ];
    
      _via_img_count -= 1;
      // END of copy of the VIA project_remove_file() function ########################


      _via_current_image_loaded = false;
      _via_buffer_img_index_list.splice( imageIndex, 1);
      _via_buffer_img_shown_timestamp.splice(imageIndex, 1);
      _via_buffer_img_index_list.splice( imageIndex, 1 )
      
      delete this.surfaces[ id ];
      editor.activeZoneList = null;
    }

    
    /**
     * Trivia.synch()
     * @summary function to synchronise the VIA and TRIVIA data
     * @version 0.2 2019-05-21 stable but dirty
     * */
    synch(){
        for( let key in this.viaObjects ){
            if (!(key in this.surfaces)) {
              let surface = new Surface( key )
              surface.addToIndex( this.surfaces )
            }
        }
        for (let key in this.surfaces){
            if ( !( key in this.viaObjects ) ) {
                delete this.surfaces[key]
            }
        }
    }

};




/***********************************************************************************
 * Trivia Surface Class
 * 
 * @description:
 * ...
 * 
 * @todo:
 * - needs a little bit of tidying 
 * - naturalHeight & naturalWidth sollen direkt beim load in die file attributes des via objects geschrieben werden
 * - imgRef & textRef in die file_attributes eines surfaces schreiben
 * - (1) rename to zoneHierarchy, (2) backend: check if i need to den regions and zones at update
 */
class Surface {
    constructor( id ) {
        this.id = id;
        this.viaObject = trivia.getViaObjectById( this.id )
        this.zoneIds = this.viaObject.regions.length
    }

    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * Surface getter
     * **************
     * - canvas (the VIA canvas)
     * - measures
     * - name
     * - image (the VIA Image)
     * - regions
     * - zones
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/  

    /**
     * Surface.canvas
     * @description the canvas of a surface
     * @returns an object
     */
    get canvas(){
      return _via_reg_canvas
    }

    /**
     * Surface.measures
     * @description the absolute height and with of a surfaces image
     * @returns an object
     */
    get measures(){
      return trivia.getSurfaceSizeByImageId( this.id )
    }

    /**
     * Surface.name
     * @description the name of a Trivia Surface
     * @returns an html <img/> dom element
     */
    get name(){
      return this.viaObject.filename
    }

    /**
     * Surface.image
     * @description the buffered image of a surface from the html dom
     * @returns an html <img/> dom element
     */
    get image(){
      return trivia.bufferedImages()[ this.id ]
    }

    /**
     * Surface.regions
     * @description function that returns the list of VIA regions of the surface.
     * @returns a list containing all VIA regions of a surface
     */
    get regions(){
      return this.viaObject.regions
    }

    /**
     * Surface.zones
     * @description function creating corresponding TRIVIA Zone instances from VIA Regions. It also sets a default id if there isn't any, since the zones editor
     *              need ids to operate. Result is an index containing all zone-ids including index from their position in the via objects regions list
     * @returns an object containing zone-ids and corresponding TRIVIA Zone instances 
     */
    get zones() {
      let zonesIndex = {}
      let regions = this.regions
      let index = 0
      for ( let i = 0; i < regions.length; i++ ){
        let region = regions[i]
        let zoneId = region['region_attributes']['zoneId']
        zoneId = zoneId == 'not_defined' || zoneId == 'not defined yet!' || zoneId == undefined || zoneId == '' ? this.newRegionId() : zoneId;
        console.log(i + ': ' + zoneId)

        let zone = new Zone( this.id, zoneId, region )
        zone.index = index
        zonesIndex[ zoneId ] = zone
        index++
      }
      return zonesIndex
    }

    /**
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     * Surface methods
     * **************
     * 
     * - addToVia
     * - addToIndex
     * - export
     * - getRegionByIndex
     * - getRegionById
     * - regionIdNumbers
     * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     **/
    
    /**
     * Surface.addToIndex()
     * @description adds a surface to a given surface index object
     * @version 0.1 2019-05-15 stable
     */
    addToIndex( index ){
      if ( !(this.id in index) ){
        index[ this.id ] = this
      } 
    }

    /**
     * Surface.export()
     * @description creates an object representation of a Surface
     * @version 0.1 2019-05-15 stable
     */
    export(){
      let obj = {
        id : this.id,
        filename : '',//this.name,
        size : this.viaObject.size,
        file_attributes : this.viaObject["file_attributes"],
        regions : this.viaObject.regions, 
        zones : this.serialize().zones
      }
      if ( !obj["file_attributes"]["naturalHeight"] || obj["file_attributes"]["naturalHeight"] == '' ) {
        obj["file_attributes"]["naturalHeight"] = this.measures.naturalHeight;
      }
      if ( !obj["file_attributes"]["naturalWidth"] || obj["file_attributes"]["naturalWidth"] == '' ) {
        obj["file_attributes"]["naturalWidth"] = this.measures.naturalWidth;
      }
      obj["file_attributes"]["image_url"] = '';
      return obj
    }

    export2(){
      let obj = {
        id : this.id,
        filename : '',//this.name,
        size : this.viaObject.size,
        file_attributes : this.viaObject["file_attributes"],
        zones : this.serialize().zones
      }
      if ( !obj["file_attributes"]["naturalHeight"] || obj["file_attributes"]["naturalHeight"] == '' ) {
        obj["file_attributes"]["naturalHeight"] = this.measures.naturalHeight;
      }
      if ( !obj["file_attributes"]["naturalWidth"] || obj["file_attributes"]["naturalWidth"] == '' ) {
        obj["file_attributes"]["naturalWidth"] = this.measures.naturalWidth;
      }
      obj["file_attributes"]["image_url"] = '';
      return obj
    }

    /**
     * Surface.getRegionByIndex()
     * @description gets a region by its index
     * @version 0.1 2019-05-15 stable
     */
    getRegionByIndex( index ){
      return this.viaObject.regions[ index ]
    }

    /**
     * Surface.getRegionById()
     * @description gets a region by its ID
     * @version 0.1 2019-05-15 stable
     */
    getRegionById( id ){
      return this.zones[ id ].viaRegion
    }

    /**
     * Surface.regionIdNumbers()
     * @description get all asigned id numbers asigned to the VIA regions
     * @returns the VIA Region object
     */
    regionIdNumbers(){
      let list = []
      for (let key in this.viaObject.regions){
        let id = this.viaObject.regions[key].region_attributes.zoneId;
        let re = /zone(\d+)/
        let extractedId = re.exec(id)
        if( extractedId ){
          list.push( parseInt(extractedId[1]) )
        }
      }
      
      return list.sort( function(a,b){
          return a - b
      })
    }

    /**
     * Surface.removeFromIndex()
     * @description removes a surface from a given surface index object
     * @version 0.1 2019-05-16 stable
     */
    removeFromIndex( index ){
      if ( this.id in index ){
        delete index[ this.id ]
      } 
    }

    /**
     * Surface.updateWithObject()
     * @description function to update the surface data object (aka. the corresponding via _via_image_metadata object) with the given json object
     * @version 0.2 2019-05-10 in_work
     * @summary runtime O(2n)
     * @todo:
     *  - cleaning up: this function iterates to often! Just iterate once and update corresponding VIA Zones
     *  - figuring out how to use the .saved status properly. Mayby during export when the TextGrid update is initialised
     *    from within the editor. so the lastActive list can be checked if saved. if not it should be updated and then pushed to textgrid
     * 
     * @returns the updated regions of a surface
     */
    updateWithObject( zoneList ){
      console.log('Surface.updateWithObject: Updating surface ' + this.id + ' with object from ZoneEditor')
      let zones = this.zones;

      function _updateRegionsFromHierarchy( zoneList ){
        for(let i=0; i < zoneList.length; i++ ){
          let zone = zoneList[i]
          let zoneId = zone.id
          let children = zone.children
          let correlatingRegion = zones[ zoneId ].viaRegion
          
          // update the chilfOf relation of all aggregated zones
          if( children ){
            
            for (let i = 0; i < children.length; i++){
              let source = children[ i ];
              let target = zones[ source.id ].viaRegion;
              target.region_attributes.childOf = zoneId;
            }

            _updateRegionsFromHierarchy( zone.children )
          }
        }
      }

      //console.log( zoneList )
      let regions = this.regions
      
      // clean all region attributes
      for (let key in regions){
        let regionAttr = regions[key]["region_attributes"]
        regionAttr.childOf = "not_defined"
      }

      // update all regions rekursively from the json
      _updateRegionsFromHierarchy( zoneList )
      
      return this.regions
    }

    /**
     * Surface.newRegionId()
     * @description function to create a new region id
     * @returns the new id
     */
    newRegionId(){
      let regionIds = this.regionIdNumbers();
      let newIdNr = regionIds.length == 0 ? 0 : regionIds[ regionIds.length - 1 ] + 1; // get the last id and add 1
      let newId = "zone" + newIdNr;  
      return newId;
    }

    /**
     * Surface.setViaRegionId()
     * @description function to set the zoneId of a VIA Region
     * @returns the VIA Region object
     */
    setViaRegionId( region ){
      region['region_attributes']['zoneId'] = this.newRegionId();
      return region;
    }

    /**
     * Surface.serialize()
     * @description: serializing a surface to a object to build a list from it etc.
     * @version 0.3 2019-05-15 stable 
     * @summary runtime O(n) depending on regions.length
     * @todo: (1) rename to zoneHierarchy
     */
    serialize(){
      let regions = this.regions
      let zones = this.zones
      let rootZones = []
      
      // iterate over Surface.regions in VIA order
      for (let i in regions){
        let region = regions[i];
        let zoneId = region['region_attributes']['zoneId'];
        let childOf = region['region_attributes'].childOf && region['region_attributes'].childOf != "not_defined" ? region['region_attributes'].childOf : null;
          
        // secure if there is no zone with this id: if this zoneid exists if not hang it on to root
        if ( childOf &&  childOf in zones ) { 
          let targetZone = zones[ childOf ];
          
          // if there is no children attribute yet create it and push the current zone into it
          if( !targetZone.children ) { 
            targetZone.children = [] 
          }

          let zone = zones[ zoneId ]
          //zone.data = zone.viaRegion()
          targetZone.children.push( zone.export() )
            
        } else {
          // if it is a root zone that is not aggregated by any othe zone than add it to rootZones, a list containig all root zones
          let zone = zones[ zoneId ]
          //zone.data = zone.viaRegion()
          rootZones.push( zone.export() )
        }
      }

      let obj = { id : this.id, name : this.name, zones : rootZones}
      return obj
    }

    /**
     * Surface.zoneList()
     * @description initializing a ZoneList for the surface in question to render a HTML list etc.
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a TriviaZoneList instance
     */
    zoneList(){
      return new TriviaZoneList( this )
    }

    /*
    zoneUI(){
      let list = [];

      function _render( zone ){
        let obj = { id : zone.id, name: zone.name, type: zone.type, index: zone.index }

        if ( !zone.children ){
          return obj
        } else {
          obj.children = [];
          for (let j = 0; j < zone.children.length; j++){
             obj.children.push( _render( zone.children[j] ) )
          }
          return obj
        }
        
      }

      let hierachy = this.serialize().zones;
      
      for(let i = 0; i < hierachy.length; i++) {
        list.push( _render(hierachy[i]) )
      } 
      
      return list
    }
    */

};  // END Class "Surface"




/***********************************************************************************
 * @class Zone
 * @description The Zone class encapsulating methods to gather data from via regions and represent them as trivia zones
 * @version 0.1 2019-05-10 stable
 * @param surfaceId the id of the surface a zone is situated on
 * @param zoneId the id of the zone
 */
class Zone {
    constructor( surfaceId, zoneId, viaRegion ) {
      this.surfaceId = surfaceId;
      this.viaRegion = viaRegion;
      this.id = zoneId;
    }

    /**
     * Zone.id Getter
     * @description gets a Zone id
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    get id() {
      return this.viaRegion.region_attributes.zoneId
    }

    /**
     * Zone.id() Setter
     * @description sets a Zone id
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    set id( id ){
      this.viaRegion.region_attributes.zoneId = id;
    }

    get name(){
      return this.viaRegion.region_attributes.name;
    }

    get type(){
      return this.viaRegion.region_attributes.mapItem;
    }

    shape(){
      return this.viaRegion[ 'shape_attributes' ]
    }

    /**
     * Zone.export()
     * @description exports a Zone
     * @version 0.1 2019-05-10 stable
     * 
     * @returns a the zone instance
     */
    export(){
     return this
    }

    

};


/***********************************************************************************
 * Trivia Zone Editor Class returning a editor object
 */
class TriviaZoneEditor {
  constructor( element, trivia ) {
      this.panel = 'trivia_zones_panel'
      this.element = $( element );
      this.lastActive = null;
      this.activeZoneList = null;
      this.trivia = trivia;
      this.init()
  }

  /**
   * TriviaZoneEditor.init()
   * @description initiate a Zone Editor instance and installs a save mutation listener on the html element #trivia_zone_pannel
   * @version 0.1 2019-05-10 stable
   * 
   * @param surfaceId the id of a surface
   * @returns a TriviaZoneList instance
   * 
   * @todo:
   *  - the panel id should be given during initialisation of the class
   */
  init(){
    console.log("TRIVIA.ZoneEditor: initiate Zone Editor instance and save mutation listener");
    let editor = this
    // start listening on trivia_zones_panel
    var targetNode = document.getElementById( editor.panel );
    var observer = new MutationObserver(function(){
      if( targetNode.attributes.class.nodeValue == 'display_area_content display_none' && editor.activeZoneList ){
        let status = editor.activeZoneList.saved
        if ( status == false ){
          console.log("TRIVIA.ZoneEditor.saveListener: Save the current zone list on " + editor.activeZoneList.id )
          editor.updateSurface( editor.activeZoneList.id )
          //editor.activeZoneList.saved = true
        } else {
          console.log("TRIVIA.ZoneEditor.saveListener: Zone list of " + editor.activeZoneList.id + " already saved!")
        }
      }
    });
    observer.observe(targetNode, { attributes: true, childList: true });
  }

  /**
   * TriviaZoneEditor.load()
   * @description function to load a zone list from a Surface instance by id
   * @version 0.1 2019-05-10 stable
   * 
   * @param surfaceId the id of a surface
   * @returns a TriviaZoneList instance
   */ 
  load( surfaceId ){
    console.log( "TRIVIA.ZoneEditor.load(): Load zone list on " + surfaceId )
    let zoneList = this.trivia.surfaces[ surfaceId ].zoneList()

    return zoneList
  }

  /**
   * TriviaZoneEditor.currentList()
   * @description function to load a zone list from the currently active Surface instance
   * @version 0.1 2019-05-10 stable
   * 
   * @returns a TriviaZoneList instance
   */ 
  currentList(){
    let activeSurfaceId = this.trivia.activeSurface.id
    console.log("TRIVIA.ZoneEditor.currentList(): Start zone editor with list on " + activeSurfaceId + ' as active list')
    this.activeZoneList = this.load( activeSurfaceId )
    return this.activeZoneList
  }

  /**
   * TriviaZoneEditor.replace()
   * @description function to replace a html zone-list with a given list used during toggeling the editor: once the eidtor is toggled the new zone list is replaced
   * 
   * @param list new list that replaces the old
   */
  replace( list ){
    this.element.replaceWith( list )
    this.element = $("ul#uinested")
  }

  /**
   * TriviaZoneEditor.serialize()
   * @description function serializing the html zone list to json
   * 
   * @returns a json string representing the html zone
   */
  serialize(){
    return serialize( this.element )
  }

  /**
   * TriviaZoneEditor.updateSurface()
   * @description function to update the last active zoneList. "updating" means the 
   *              process of saving the current state of the last active list to the the surface instance.
   * @version 0.1 2019-05-10 in_work
   * @todo:
   *  - cleaning up and figuring out how to use the .saved status properly. Maybe during export when the TextGrid update is initialised
   *    from within the editor. so the lastActive list can be checked if saved. if not it should be updated and then pushed to textgrid
   */
  updateSurface( surfaceId ){
    console.log("TRIVIA.ZoneEditor.updateSurface(): Attempting to update surface " + surfaceId)
    let surface = trivia.surfaces[ surfaceId ];
    let activeZoneList = this.activeZoneList
    let serializedList = this.serialize();
    if ( surface && serializedList ){
      console.log("TRIVIA.ZoneEditor.updateSurface(): Updateing surface " + surfaceId)
      activeZoneList.saved = true;
      return surface.updateWithObject( serializedList );
    } else {
      console.log("TRIVIA.ZoneEditor.updateSurface(): ERROR on updating " + surfaceId)
    }
  }

  /**
   * TriviaZoneEditor.toggle()
   * @description toggleing function to toggle the html zone editor. It takes the currently active surface and its zoneList, 
   *              replaces the 'old' list with the new one and renders it to HTML. Then it displays the zoneEditor.
   * @verion 0.1 2019-05-10 stable
   */
  toggle() {
    // load editor from currently avtive surface (aka. the image that is shown)
    
    let currentList = this.currentList();
    let renderedList = currentList.render();
    $('#zones-count').html( '<span class="ids">' + currentList.surface.regions.length + '</span>' );
    // load editor into div
    this.replace( renderedList )
    set_display_area_content(VIA_DISPLAY_AREA_CONTENT_NAME.ZONES)
  };

  zoneUI( surface ){
    let list = [];

    function _render( zone ){
      
      let obj = { id : zone.id, type : zone.type, index : zone.index }
      obj.name = zone.name == 'not_defined' ? '<'+zone.id+' />' : zone.name;

      if ( !zone.children ){
        return obj
      } else {
        obj.children = [];
        for (let j = 0; j < zone.children.length; j++){
           obj.children.push( _render( zone.children[j] ) )
        }
        return obj
      }
      
    }

    let hierachy = surface.serialize().zones;
    
    for(let i = 0; i < hierachy.length; i++) {
      list.push( _render(hierachy[i]) )
    } 
    
    let zoneUI = new ZoneUI( '#tree1', list );
    
    return zoneUI.render({
      autoOpen: true,
      dragAndDrop: true
    })
  }

};


class ZoneUI {
  constructor( selector, data ) {
    this.data = data;
    this.$UI = $( selector ); 
  }

  render( config ) {
    this.tree = this.$UI.tree( Object.assign(config, {data : this.data} ) );
    return this;
  }

  toJson(){
    return this.tree.tree('toJson')
  }

}

/***********************************************************************************
 * TriviaZoneList
 * @description Element Class constructing the zone lists used in the zone editor
 */
class TriviaZoneList {
  constructor( surface ) {
      this.surface = surface;
      this.saved = false;
      this.id = surface.id;
      this.element = $( '<ul id="uinested" class="uk-nestable" data-uk-nestable="" data-id="'+this.id+'"></ul>' );
      TriviaZoneList.prototype.instances++;
  }

  /**
   * TriviaZoneList.render()
   * @description rendering function to render a trivia HTML zone-list
   */
  render(){
    
    /**
     * TriviaZoneList._renderItem()
     * @description recursive function to render TriviaZone Listitems
     * @param {*} zone 
     */
    function _renderItem( zone ){
      // get id, name, shape
      let id = zone.id
      let name = zone.viaRegion["region_attributes"].name
      let type = zone.viaRegion["region_attributes"].mapItem
      let shape = zone.viaRegion["shape_attributes"].name
      let index = zone.index
      
      if ( zone.children && zone.children.length > 0 ){
        // create a complex item containing a sublist
        let complexItem = new TriviaListItem(id, name, shape, type, index).aggregated()
        // create the sublist
        let sublist = $( '<ul class="uk-nestable-list"></ul>' )
        // ... and now append all subitems recursively
        for(let item in zone.children){
          sublist.append( _renderItem( zone.children[item] ) ) 
        }
        //return a complex iter
        return complexItem.append( sublist )
      } else {
        // return a simple list item
        return new TriviaListItem(id, name, shape, type, index).standard()
      }
    }
    
    // get all the zones form the surface
    let zones = this.surface.serialize().zones
    
    // now render all zones rekursively
    for (let key in zones){
      //console.log("render list items rekursively")
      let zone = zones[key]
      let listItem = _renderItem( zone )
      this.element.append( listItem )
    }
    // return the constructed html list
    return this.element
  }

  /**
   * TriviaZoneList.serialize()
   * @description function serialising a html trivia zone list to a json representation
   */
  serialize() {
    let triviaList = this.render()
    return serialize( triviaList ) 
  };

}
TriviaZoneList.prototype.instances=0;



/***********************************************************************************
 * Element Class constructing the items of a zone list
 */
class TriviaListItem {
  constructor( id, name, shape, type, index ) {
      this.id = id;
      this.name = name;
      this.shape = shape;
      this.index = index;
      this.type = type;
  }

  /**
   * TriviaListItem.createDefaultName()
   * @description function creating a default name displayed in the list item
   * @returns a name as string
   */
  createName(){
    let name = !this.name || this.name == "undefined" || this.name == 'not_defined' ? '<span class="zone-id">'+this.id+'</span>' : '<span class="name">'+this.name+'</span>';
    name += this.type ? '<span class="type">' + this.type + '</span>' : '';
    return name
  }

  /**
   * TriviaListItem.standard()
   * @description function rendering a standard zone-list item that has no children
   * @returns a jquery constructor
   */
  standard() {
      this.element = $('<li class="uk-nestable-item trivia-zone" data-shape="'+this.shape+'" data-id="'+this.id+'" data-index="'+this.index+'" data-name="'+this.name+'"><div class="uk-nestable-panel">' + this.createName() + '</div></li>');
      return this.element;
  }

  /**
   * TriviaListItem.aggregated()
   * @description function rendering a complex zone-list item that has children aka. children other zones
   * @returns a jquery constructor
   */
  aggregated() {
      this.element = $('<li class="uk-nestable-item uk-parent trivia-zone" data-shape="'+this.shape+'" data-id="'+this.id+'" data-index="'+this.index+'" data-name="'+this.name+'"><div class="uk-nestable-panel">' + this.createName() + '</div></li>');
      return this.element
  }
}
  


/***********************************************************************************
 * TextGrid Class
 * 
 * testuris: textgrid:3r346, textgrid:3r3p5, textgrid:3r527
 */
class TriviaTextGridClient {
  constructor( config ) {
    this.config = config;
    this.resources = {};
  }

  /**
   * TriviaTextGridClient.activeResource()
   * @description function to get the currently active resource
   * @returns the currently active resource 
   */
  activeResource(){
    if ( this.resources.length > 0 && this.trivia ){
      return this.resources[ _via_image_id ]
    }
  }

  /**
   * TriviaTextGridClient.init()
   * @description init function
   */
  init( panel ){
    let client = this;
    client.panel = $( panel );
    client.resourceList = this.panel.find('#textgrid_resources_list > ul');

    if( client.config == undefined ){
      client.getConfigFromTrivia()
        .done(function( data ){
          console.log("TriviaTextGridClient.init: Credentials loaded from trivia backend");
          client.config = data;
          client.updatePanel();
        })
        .fail( function(){
          console.log("TriviaTextGridClient.init: Error on loading credentials from trivia backend");
          console.log( err );
        })
    } else {
      console.log("TriviaTextGridClient.init: client not configured");
    }
  }

  /**
   * TriviaTextGridClient.updatePanel()
   * @description function tu update the textgrid project panel 
   */
  updatePanel(){
    this.panel.find( '#trivia-textgrid-project' ).val( this.config.pid );
  }

  /**
   * TriviaTextGridClient.createEntry()
   * @description function to create an entry in textgrid panel list
   * @returns a DOM element 
   */
  createEntry( resource ){
    let resourceList = this.resourceList;
    resourceList.find(".sel").removeClass("sel")
    let listItem = $('<li id="tg'+ (_via_img_count-1) +'" class="trivia-resource-list-item sel" title="Resource Name" data-uri="' + resource.uri + '" data-id="' + resource.uri + '"></li>')
    let title =  $('<span class="trivia-resource-title" onclick="tgclient.jumpToImage(this, '+ (_via_img_count-1) +')">' + resource.title + '</span>')
    let icon = $('<span class="trivia-resource-config-icon" onclick="configTextGridTriviaResource(this)"><svg viewbox="0 0 24 24"><use xlink:href="#icon_settings"></use><title>Configure</title></svg></span>')
    let update = $(' <span class="trivia-resource-config-icon" onclick="tgclient.updateTriviaResource(\'' + resource.uri + '\')"><svg viewbox="0 0 24 24"><use xlink:href="#icon_save"></use><title>Update Resource within TextGrid</title></svg></span>')
    listItem.append(title);
    listItem.append(icon);
    listItem.append(update);
    resourceList.append( listItem )
    return listItem
  };

  loading( action ){
    let display = action == 'start' ? 'inline' : 'none';
    $( '.loading-gif' ).css( 'display', display)
  }
  
  /**
   * TriviaTextGridClient.getConfigFromTrivia()
   * @description function to load configured data from the backend
   */
  getConfigFromTrivia(){
    return $.ajax({
      url: trivia.backend + "restxq/trivia/1.1/auth",
      type: 'GET',
      crossDomain: true,
      //dataType : "json"
    });
  }

  testFetch(){
    fetch( trivia.backend +  'restxq/trivia/1.1/auth')
            .then( data => {return data.json()})
            .then( res => {console.log(res)})
  }

  /**
   * TriviaTextGridClient.loadTriviaResource()
   * @description function to load a trivia resource from textgrid and backend into trivia frontend using a via user input form.
   *              This is the main interface to load trivia resources into the frontend
   * @version 0.1 2019-05-10 working 
   */
  loadTriviaResource(){
    console.log('TriviaTextGridClient.loadTriviaResource: Loading trivia resource into trivia frontend')
    
    let client = this;
    let config = { 'title':'Load Trivia Resource from TextGrid' };
    let input = { 
      'textGridUri': { type:'text', name:'TextGrid URI of a Trivia Resource', placeholder:'textgrid:3r3p5', disabled:false, size:50 },
      };
  
    let loadHandle = function(){
      if ( input.textGridUri.value !== '' ) {
        let uri = input.textGridUri.value.trim();
        console.log('TriviaTextGridClient.loadTriviaResource.loadHandle: Retriving ' + uri + ' from textgrid ...')
        
        let triviaResource = new TriviaResource( uri )
        triviaResource.set( client.config )
        //triviaResource.register( client )
        client.loading( 'start' )
        
        triviaResource.ingest()
          .done( function() { 
            client.loading( 'stop' )
            triviaResource.listElement = client.createEntry( triviaResource )
            triviaResource.available = true;
            client.resources[ triviaResource.uri ] = triviaResource
            show_message('Added TextGrid resource ' + triviaResource.uri + ' with image resource [' + triviaResource.imageRef + ']');
          })
          .fail( function(){
            console.log( "TriviaTextGridClient.loadTriviaResource.loadHandle: Failed to load Trivia Resource" )
            show_message( 'Error: Failed to load TextGrid resource ' + triviaResource.uri );
            triviaResource.available = false;
            client.loading( 'stop' )
          })
      }
    };

    invoke_with_user_inputs( loadHandle, input, config );
  }

  /**
   * TriviaTextGridClient.createTriviaResource()
   * @description function to create a trivia resource within textgrid using a via user input form.
   *              This is the main interface to create trivia resources in textgrid
   * @version 0.1 2019-05-10 in development 
   * @todo:
   *  - if image uri is not in textgrid dont create > call tgcrud for image metadata in frontend
   */
  createTriviaResource(){
    console.log('TriviaTextGridClient.createTriviaResource: Creating trivia resource within textgrid')
    
    let client = this;
    let config = { 'title':'Create Trivia Resource from TextGrid' };
    let input = { 
      'title'   : { type:'text', name:'Title of Resource', placeholder:'The Title of the created Trivia Resource eg. trivia resource' +  Date.now(), disabled:false, size:50 },
      'imageUri': { type:'text', name:'Image Resource', placeholder:' The textgrid uri of the image resource eg. textgrid:3r3hn', disabled:false, size:50 },
      'textUri': { type:'text', name:'TEI Resource', placeholder:' The textgrid uri of an associated TEI resource eg. textgrid:3r3hn', disabled:false, size:50 },
      };
  
    let createHandle = function(){
      let title = input.title.value.trim();
      let imageUri = input.imageUri.value.trim();
      let textUri = input.textUri.value.trim();

      if ( title !== '' && imageUri !== '' ) {
        console.log('TriviaTextGridClient.createTriviaResource: Creating resource [' + title + '] in textgrid');
        let triviaResource = new TriviaResource()
        //triviaResource.register( client )
        triviaResource.set( client.config )
        client.loading( 'start' )
        
        triviaResource.create( title, imageUri, textUri )
          .done( function(){ 
            triviaResource.ingest()
              .done( function () {
                triviaResource.listElement = client.createEntry( triviaResource )
                triviaResource.available = true;
                client.resources[ triviaResource.uri ] = triviaResource
                client.loading( 'stop' )
                show_message('Created and Added TextGrid resource ' + triviaResource.uri + ' with image resource [' + triviaResource.imageRef + ']');
              })
          })
          .fail( function(){
            console.log( "TriviaTextGridClient.createTriviaResource.createHandle: Failed to create Trivia Resource" )
            show_message( 'Error: Failed to create new TextGrid resource!' );
            triviaResource.available = false;
            client.loading( 'stop' )
          })
      } else {
        console.log('TriviaTextGridClient.createTriviaResource: Could not create Trivia Resource without Title or Image URI!');
        show_message('Error: Could not create Trivia Resource without Title or Image URI!!');
      }
    };

    invoke_with_user_inputs( createHandle, input, config );
  }

  /**
   * TriviaTextGridClient.removeTriviaResource()
   * @description function to remove a trivia resource from the frontend
   *              This is the main interface to remove trivia resources
   * @version 0.1 2019-05-10 development 
   * @todo:
   *  - implement functionality
   */
  removeTriviaResource(){
    console.log('TriviaTextGridClient.removeTriviaResource: Removing trivia resource from trivia frontend')
    let id = trivia.activeSurface.id;
    trivia.removeSurfaceFromTrivia( id );
    // not nice but handy to load the first image off all
    if ( Object.keys(_via_img_metadata).length > 0 ){
      _via_show_img(0)
    } else {
      set_display_area_content(VIA_DISPLAY_AREA_CONTENT_NAME.PAGE_GETTING_STARTED);
    }
    
    $( this.resources[ id ].listElement ).remove()
    show_message('Deleted TextGrid resource ' + id + ' from TriVIA');
  }

  /**
   * TriviaTextGridClient.updateTriviaResource()
   * @description function to update a trivia resource in TextGrid
   *              This is the main interface to update trivia resources
   * @version 0.1 2019-05-10 development 
   * @todo:
   *  - implement functionality
   */
  updateTriviaResource( uri ){
    console.log('TriviaTextGridClient.updateTriviaResource: Updateing Trivia Resource in TextGrid')
    let client = this
    client.loading( 'start' )
    
    
    //console.log( editor.activeZoneList.saved );
    if( editor.activeZoneList && editor.activeZoneList.saved != true ){
      console.log("Not saved yet! needed implementation");
      editor.updateSurface( uri );
      // update current editor
    }

    if ( trivia.world == 'local') {
      console.log('xml request update');
      this.resources[ uri ].update();
    } else { // AJAX does not work yet !!
      console.log('ajax update')
      this.resources[ uri ].update()
        .done( function( data ) {
          console.log( "Ajax update successfull" );
          console.log( data );
          client.loading( 'stop' )
          show_message('Updated TextGrid Resource ' + uri);
        })
        .fail( function( err ) {
          console.log( "Error: Ajax update failed" );
          console.log( err )
          client.loading( 'stop' )
          show_message('Error on updating TextGrid Resource ' + uri);
        });
    }
  }

  jumpToImage( element, index ){
    $("#textgrid_resources_list > ul").children().removeClass("sel")
    let parent = $( element.parentElement ).addClass('sel');
    _via_show_img(index);
    //jump_to_image(index)
  }
    
};

/***********************************************************************************
 * TextGridResource
 */
class TextGridResource{
  constructor( uri, sid, pid, crud = 'https://textgridlab.org/1.0/tgcrud/rest' ){
    this.uri = uri;
    this.sid = sid;
    this.pid = pid;
    this.crud = crud;
  }

  /**
   * TextGridResource.register()
   * @description registering the TextGridResource instance in the given TriviaTextGridClient Instance and loading config
   * @param tgclien the TriviaTextGridClient Instance this resource should be registerd in
   * @returns itself (instance of TextGrid Resource)
   */
  register( tgclient ){
    this.set( tgclient.config )
    tgclient.resources[ this.uri ] = this
    return this
  }

  get( uri ){
    console.log( "get metadata and data from textgrid" )
    // nearly the same functionailty as in trivia resource
  };


  /**
   * TextGridResource.set()
   * @description setting the parameters
   * @param config 
   * @todo:
   *  - delete the hard coded credentials and load from somewhere else maybe via tgclient
   */
  set( config ){
    this.pid = config['pid'];
    this.sid   =   config['sid'],
    this.crud  =   config['tgCrud'],
    this.aggregations   =   config['aggregations']
    return this
  }


  /**
   * TextGridResource.meta()
   * @description getting metadata of a textgrid resource from tgcrud
   * @returns an ajax object
   * @todo:
   */
  meta(){
    return $.ajax({
      url: this.crud + '/' + this.uri + '/metadata?sessionId=' + this.sid,
      type: 'GET',
      crossDomain: true,
      contentType: "json",
    })
  }


  /**
   * TextGridResource.data()
   * @description getting a textgrid resource from tgcrud
   * @returns an ajax object
   * @todo:
   */
  data(){
    return $.ajax({
      url: this.crud + '/' + this.uri + '/data?sessionId=' + this.sid,
      type: 'GET',
      crossDomain: true,
      contentType: "json",
    })
  }


};

/***********************************************************************************
 * TriviaResource < TextGridResource
 */
class TriviaResource extends TextGridResource{
  constructor( uri, sid, pid, crud ){
    super(uri, sid, pid, crud)
    this.triviaCrud = trivia.backend + "restxq/trivia/1.1/textgrid/crud";
    this.textRef = [];
    this.imageRef = null;
  };

  /**
   * TriviaResource.create()
   * @summary function to create a trivia resource in textgrid
   * @version 0.1 2019-05-10 in progress
   * @returns a ajax object
   * @todo:
   *  - this function is quick and dirty ... so tidy up!
   * */
  create( title, imageRef, textUri ){
    console.log("TriviaResource.create: create trivia resource")
    let url = this.triviaCrud + '/create?sid=' + this.sid + '&pid=' + this.pid 
    url += '&imageUri=' + imageRef;
    url += textUri ? '&textUri=' + textUri : '';
    url += '&title=' + title;
    url += this.aggregations.trivia ? '&aggregation=' + this.aggregations.trivia : '';

    console.log( encodeURI(url) )
    
    let create = $.ajax({
      url: encodeURI(url),
      type: 'GET',
      crossDomain: true,
      //contentType: "json",
    })

    let instance = this 
    return create.then( function( data ){
      instance.uri = data.created.triviaResource.uri;
      instance.textRef = data.created.teiResource ? data.created.teiResource.uri : ''
    })
    
  }

  
  /**
   * TriviaResource.data()
   * @summary function to get a converted trivia resource from textgrid via trivia. This function is a trivia wrapper around the textgrid http GET data method.
   * 
   * @returns an jquery ajax object that may be chained
   * */
  data(){
    return $.ajax({
      url: this.triviaCrud + '/' + this.uri + '/data?sid=' + this.sid,
      type: 'GET',
      crossDomain: true,
      //contentType: "json",
    })
  }


  /**
   * TriviaResource.get()
   * @summary function to load data from textgrid into the Trivia Resource instance
   * 
   * @returns a jquery ajax object containing the finished get request
   *  */  
  get(){
    console.log( "TriviaResource.get: get converted recource from textgrid via trivia api" )
    let instance = this

    let metadataObject = this.meta();
    let dataObject  = this.data();
    
    // fire metadata request and then ...
    return metadataObject.then(
      function( metadata ){
        console.log("TriviaResource.get: get title from TextGrid");
        instance.title = metadata.getElementsByTagName("title")[0].innerHTML;
        console.log( metadata );

        // ... fire data request to exist-db backend to retrieve converted trivia resource
        return dataObject.done( function( data ){
          //dataCallback( data, instance );
          console.log("TriviaResource.get: get data from TextGrid"); 
          console.log(data);
          instance.object = data;
          instance.imageRef = data.file_attributes.imageRef 
          instance.textRef = data.file_attributes.textRef 
      
      // set image_url in filename and file_attributes.image_url
      let imageUrl = instance.imageUrl();
      instance.object.filename = imageUrl;
      instance.object.file_attributes.image_url = imageUrl;
        })
        .fail( function() {
          console.log( "TriviaResource.get.data: error loading TRIVIA Resource")
        })

      }
    ).fail( function() {
      console.log( "TriviaResource.get.metadata: error loading TRIVIA Resource")
    })
  };


  /**
   * TriviaResource.imageUrl()
   * @summary function to construct a textgrid url for a image resource
   * 
   * @returns an url as string
   * */
  imageUrl( imageUrl ){
    let uri = imageUrl ? imageUrl : this.imageRef
    let url = this.crud + "/" + uri + '/data'
    url += this.sid ? '?sessionId=' + this.sid : ""
    return url 
  }


  /**
   * TriviaResource.ingest()
   * @summary function to "ingest" aka. load data from textgrid and trivia backend to trivia frontend.
   *          It loads the data into the frontend data- and html structure.
   * @todo bind it to surface as method
   * @version 0.1 2019-05-10 stable
   * */
  ingest(){
    console.log("TriviaResource.ingest: Loading trivia resource from textgrid into trivia application");
    
    let instance = this;
    return this.get().done( function(){
      console.log('TriviaResource.ingest: retrived data for ' + instance.uri + ' from TextGrid. Ingesting into Trivia'); 
      trivia.addSurfaceToTrivia( instance.uri, instance.object )
        let img_index = _via_image_id_list.indexOf( instance.uri );
        _via_show_img(img_index);
        trivia.synch();
      })
  };

  update( ){
    console.log( 'TRIVIA.TriviaResource.update: Updating trivia resource ' + this.uri + ' within textgrid ...');
    let uri = this.uri;
    
    // check if resource is actually represented in surfaces
    if ( trivia.surfaces[ uri ] ) {
      this.object = trivia.surfaces[ uri ].export();
      
      //console.log( this )
      let postData = JSON.stringify( this.object );
      let url = this.triviaCrud + '/' + this.uri + '/update?sid=' + this.sid;
      //let url = 'https://127.0.1.1:8443/exist/restxq/trivia/1.1/textgrid/crud/' + this.uri + '/update?sid=' + this.sid;
      
      console.log( postData );
      console.log( url );  

      if ( trivia.world == 'local'){
        var req = new XMLHttpRequest();
        // Cross Origin for testing purposes
        req.open('POST', url, true);
        //req.setRequestHeader('Content-Type', 'application/json');
        req.onreadystatechange = function() {
          if (req.readyState === 4) {
              if (req.status >= 200 && req.status < 400) {
                console.log("XMLRequest done")
                //console.log(req.responseText);
                tgclient.loading( 'stop' )
                show_message('Updated TextGrid Resource ' + uri);
                //return req.responseText;
              } else {
                console.log( "Error: XMLRequest update failed" );
                show_message('Error on updating TextGrid Resource ' + uri);
                tgclient.loading( 'stop' )
                console.log(req)
              }
            }
          };
        req.send( postData );

      } else {
        return $.ajax({
          url: url,
          type: 'POST',
          crossDomain: true,
          data: postData
          //contentType: "json",
        })
      }
      
    } else {
      console.log('TRIVIA.TriviaResource.update: Error since ' + uri + ' is not a surface!' )
    }
  }

};



//***************************************
//***** General Functions
//**


function serialize( list ) {
    
    function walkListItem( listItem ) {
        // collect data attributes and put them into an object. this behavour is the final case
        let objData = listItem.dataset
        let obj = { "name" : objData["name"], "id" : objData["id"], "index" : objData["index"] }
        let subList = $(listItem).children("ul")
        if ( subList.length > 0 ) {
            // recursive case if there are further lists. an children property is created
            obj["children"] = new Array()
            
            $(subList).children("li").each(function () {    
                obj["children"].push( walkListItem( this ) )
            });
        }
        return obj
    };

    let objects = new Array();

    // run through the list and collect all objects
    let node = document.getElementById( list );
    $( list ).children("li").each(function () {    
        objects.push( walkListItem( this ) )
    });
    
    return objects
};



function configTextGridTriviaResource( element ){
  let parent = element.parentElement
  let uri = parent.uri

  var config = {'title':'Configure the TextGrid TRIVIA Resource Relations' };
  var input = { 
    'trivia-uri': { type:'text', name:'TRIVIA TextGrid Resource (textgrid URI)', placeholder:'e.g. textgrid:1234', disabled:false, size:50 },
    'tei-uri': { type:'text', name:'TEI TextGrid Resource (textgrid URI)', placeholder:'e.g. textgrid:1234', disabled:false, size:50 },
  };
  
  invoke_with_user_inputs(function(){
    console.log(input)
  }, input, config);
}


/**################################################################################
 *  START UP
 *################################################################################*/
let trivia = new Trivia()
let editor = new TriviaZoneEditor( "ul#uinested", trivia );
let tgclient = new TriviaTextGridClient( );
tgclient.getConfigFromTrivia()
tgclient.trivia = trivia;

/**################################################################################
 *  VIA Mods
 *################################################################################*/


 //  loading VIA Submodules and define trivia datamodel
function _via_load_submodules() {
  _trivia_define_attributes();
}

// Wrap VIA function "project_open_parse_json_file"
/**
* if(_via_image_id_list.length > 0){ 
    for (let i = 0; i <_via_image_id_list.length; i++) {
      _via_image_id_list.pop(i)
    }
  };
*/
//@@MOD: Function mod with wrapper to assign array listener
/*
var modProjectOpen = function(fn){
  return function(){
    try{
      let funcReturn = fn.apply(this, arguments);
      
      // secondary push listener
      arrayPushEvent(_via_image_id_list, function(updatedArr) {
        let last = updatedArr[ updatedArr.length - 1 ];
        console.log("push listener 2");
        trivia.synch()
      });

      // secondary pop handler
      arrayPopEvent(_via_image_id_list, function(updatedArr) {
        console.log("pop listener 2");
        trivia.synch()
      });

      return funcReturn;
    }catch(ex){
      //ErrorHandler.Exception(ex);
      console.log("error on secondary listener")
    }
  };
};

//var project_open_parse_json_file = modProjectOpen(project_open_parse_json_file);
*/

var modZoneEditor = function( fn , trivia ){
  return function(){
    try{
      
      // build dropdown on attribute childId
      let options = {}
      let regions = trivia.activeSurface.regions;
      let selectedRegionId = _via_user_sel_region_id;
      for (let id = 0; id < regions.length; id++){
        
        if( id != selectedRegionId ){
          let region = regions[ id ];
          let count = id + 1;
          //console.log(region)
          let zoneId = region.region_attributes.zoneId;
          let name = region.region_attributes.name;
          let string = name ? name + ' (' + zoneId + ')' : zoneId;
          options[ zoneId ] = count + ') ' + string;
        }
      }
      options[ 'not_defined' ] = '0) image';
      _via_attributes['region']['childOf'].type = 'dropdown';
      _via_attributes['region']['childOf'].options = options;


      let funcReturn = fn.apply(fn, arguments);

      /*
      $("#annotation_editor").find( $("textarea")[1] ).change( function(){ 
        console.log( trivia.activeSurface().regionIds() )
      })*/
      
    }catch(ex){
      //ErrorHandler.Exception(ex);
      console.log("error on secondary listener")
      console.log( ex )
    }
  };
}
var annotation_editor_show = modZoneEditor( annotation_editor_show, trivia)


let setSid = function( sid ){
  alert( sid );
}

let addTGObject = function(uri, titleUri, contentType){
  alert( uri + titleUri, contentType);
}


function getDataFromTrivia(){
  
  var $tree = $('#tree1');
  console.log(trivia.activeSurface.zoneUI())
  $tree.tree({
      data: trivia.activeSurface.zoneUI(),
      autoOpen: true,
      dragAndDrop: true
    });
}

/**################################################################################
 *  DOCUMENT READY
 *################################################################################*/
$( document ).ready(function() {
    
    trivia.init();
    tgclient.init( '#textgrid_panel' );

    
    $( ".json-button" ).click( function() {
        let json = JSON.stringify( serialize("ul#uinested") );
        let area = document.getElementById( "json" );
        area.innerHTML = json;
    });


    /**
     * TextGridList Trigger
     */
    $( ".trivia-resource-list-item" ).click( function() {
      $(this.parentElement).children().each( function() {
        $(this).removeClass( 'sel' );
      })
      $(this).addClass( 'sel' );
    });
    
    
});


